# -*- coding: utf-8 -*-
import pandas as pd
import numpy as np

from .basic import CROSS
from ..core.indicator_signal import register_signal_indicator

# Simple Moving Average (SMA)
# --------------------------------------------
def SMA(data, length):
    return data.rolling(length, min_periods=1).mean()

# Exponential Moving Average (EMA)
# --------------------------------------------
def EMA(data, length):
    return data.ewm(span=length, min_periods=1).mean()

# Rolling Moving Average
# --------------------------------------------
def RMA(data, length):
    alpha = (1.0 / length) if length > 0 else 0.5
    return data.ewm(alpha=alpha, min_periods=1).mean()

# Weighted Moving Average
# --------------------------------------------
def WMA(data, length=None, asc=None):
    asc = asc if asc else True

    total_weight = 0.5 * length * (length + 1)
    weights_ = pd.Series(np.arange(1, length + 1))
    weights = weights_ if asc else weights_[::-1]

    def linear(w):
        def _compute(x):
            diff = len(w) - len(x)
            if diff > 0:
                x = np.insert(x, 0, [ x[0] ] * diff)
            return np.dot(x, w) / total_weight
        return _compute

    close_ = data.rolling(length, min_periods=1)
    wma = close_.apply(linear(weights), raw=True)
    return wma

# Hull Moving Average，HMA
# --------------------------------------------
def HMA(data, length):
    half_length = int(length / 2)
    sqrt_length = int(np.sqrt(length))
    wmaf = WMA(data, length=half_length)
    wmas = WMA(data, length=length)
    hma = WMA(2 * wmaf - wmas, length=sqrt_length)
    return hma

# MA, call other implmentations
# --------------------------------------------
def MA(name:str = None, data:pd.Series = None, **kwargs) -> pd.Series:
    _mas = [
        "dema", "ema", "fwma", "hma", "linreg", "midpoint", "pwma", "rma",
        "sinwma", "sma", "swma", "t3", "tema", "trima", "vidya", "wma", "zlma"
    ]
    if name is None and data is None:
        return _mas
    elif isinstance(name, str) and name.lower() in _mas:
        name = name.lower()
    else: # "ema"
        name = _mas[1]

    if name == "sma": return SMA(data, **kwargs)
    elif name == "rma": return RMA(data, **kwargs)
    elif name == "hma": return HMA(data, **kwargs)
    elif name == "wma": return WMA(data, **kwargs)
    #elif name == "dema": return dema(data, **kwargs)
    #elif name == "fwma": return fwma(data, **kwargs)
    #elif name == "linreg": return linreg(data, **kwargs)
    #elif name == "midpoint": return midpoint(data, **kwargs)
    #elif name == "pwma": return pwma(data, **kwargs)
    #elif name == "sinwma": return sinwma(data, **kwargs)
    #elif name == "swma": return swma(data, **kwargs)
    #elif name == "t3": return t3(data, **kwargs)
    #elif name == "tema": return tema(data, **kwargs)
    #elif name == "trima": return trima(data, **kwargs)
    #elif name == "vidya": return vidya(data, **kwargs)
    #elif name == "zlma": return zlma(data, **kwargs)
    else: return EMA(data, **kwargs)

def signal_ma(df, inplace=False, fast=5, slow=20):
    periods = [5, 10, 20, 30, 60]
    if not fast in periods:
        periods.append(fast)
    if not slow in periods:
        periods.append(fast)
    for period in periods:
        ma = SMA(df.close, period)
        if inplace:
            df[ 'ma' + str(period) ] = ma
        if period == fast:
            ma_fast = ma
        elif period == slow:
            ma_slow = ma
    return CROSS(ma_fast, ma_slow)

register_signal_indicator('ma', signal_ma, ['ma5', 'ma10', 'ma20', 'ma30', 'ma60'], 'MA', 'trend')



# -*- coding: utf-8; py-indent-offset:4 -*-

import datetime as dt
import smtplib
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText

from .push_base import *

class EmailPush( PushBase ):
    mailto = ''
    sender = ''
    server = ''
    user = ''
    passwd = ''

    def __init__(self, conf):
        super().__init__(conf)

        self.mailto = conf['mailto']
        self.sender = conf['sender']
        self.server = conf['server']
        self.user = conf['user']
        self.passwd = conf['passwd']

    def send(self, subject = 'HELLO', content = 'hello, world'):
        if not self.mailto:
            print('Error: email configuration missing.')
            return

        msg = MIMEMultipart("alternative")
        msg['From'] = self.sender
        msg['To'] = self.mailto
        msg['Subject'] = subject
        msg.attach( MIMEText( content, 'plain', 'utf-8' ) )

        s = smtplib.SMTP(self.server)
        if len(self.user)>0 and len(self.passwd)>0:
            s.login(self.user, self.passwd)

        s.sendmail(self.sender, self.mailto, msg.as_string())
        s.quit()

    def flush(self):
        if not self.msg_queue:
            return

        subject = '[Hiquant] ' + self.msg_queue[0]

        line = '-' * 40
        content = '\n\n'.join(self.msg_queue)
        content += '\n' + line + '\n' + dt.datetime.now().strftime('%Y-%m-%d %H:%M:%S') + '\n'

        self.send(subject, content)
        if self.verbose:
            print('Email sent')
            print(subject, '\n', content)

        self.msg_queue = []



# -*- coding: utf-8 -*-

from .basic import CROSS
from .ma import EMA
from ..core.indicator_signal import register_signal_indicator

# Moving Average Convergence and Divergence, by Geral Appel, 1979
# ----------------------------------------------------------------
def MACD(close, fast = 12, slow = 26, signal = 9):
    dif = EMA(close, fast) - EMA(close, slow)
    dea = EMA(dif, signal)
    macd = (dif - dea) * 2.0
    return dif, dea, macd

def signal_macd(df, inplace=False, fast=12, slow=26, signal=9):
    #dif, dea, macd_hist = talib.MACD(df.close, fastperiod=12, slowperiod=26, signalperiod=9)
    #if inplace:
    #    df['macd_dif'], df['macd_dea'], df['macd_hist'] = dif, dea, macd_hist
    #return pd.Series( CROSS(dif, dea), index=df.index )
    dif, dea, macd_hist = MACD(df.close, fast=fast, slow=slow, signal=signal)
    if inplace:
        df['macd_dif'], df['macd_dea'], df['macd_hist'] = dif, dea, macd_hist
    return CROSS(dif, dea)

register_signal_indicator('macd', signal_macd, ['macd_dif', 'macd_dea', 'macd_hist'], 'MACD', 'trend')



# -*- coding: utf-8; py-indent-offset:4 -*-

def parse_params_options(argv):
    params = []
    options = []
    for i in range(1, len(argv)):
        str = argv[i]
        if str[0] == '-':
            options.append(str)
        else:
            params.append(str)

    return params, options

def parse_arg(argv, required, defaults, syntax_tips):
    params, options = parse_params_options(argv)

    n = len(params)
    if n < required:
        print(syntax_tips.replace('__argv0__', argv[0]))
        exit()

    vals = list(defaults.values())
    m = len(vals)
    if n < m:
        params = params + vals[n:]

    return params[0:m], options

def dict_from_config_items(items, verbose = False):
    mapping = {}
    for k, v in items:
        mapping[ k ] = v
        if verbose:
            print(k, '=', v)
    return mapping



# -*- coding: utf-8; py-indent-offset:4 -*-

import os
import importlib.util

from ..utils import *
from .agent_simulated import *
from .agent_human import *

class Fund:
    market = None
    trader = None
    fund_id = None

    conf = {}
    fund_name = 'No Name'
    strategy = None
    agent = None
    stat_df = None

    verbose = False

    def __init__(self, market, trader, fund_id, fund_conf):
        self.market = market
        self.trader = trader
        self.fund_id = fund_id
        self.conf = fund_conf

    def set_verbose(self, verbose = True):
        self.verbose = verbose
        if self.agent is not None:
            self.agent.set_verbose(verbose)

    def set_agent(self, agent, order_cost):
        self.agent = agent
        self.agent.order_cost = order_cost

    def load_module(self, strategy_file):
        mod_name = os.path.basename(strategy_file).replace('.py', '')
        mod_filepath = os.path.abspath(strategy_file)
        if self.verbose:
            print('dynamically loading strategy:', mod_filepath)
        spec = importlib.util.spec_from_file_location(mod_name, mod_filepath)
        module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(module)
        return module

    def load_strategy(self, strategy_file):
        self.strategy_file = strategy_file

        mod = self.load_module( strategy_file )
        mod.init( self )

    def init_fund(self, realtime_trade = False):
        self.fund_name = self.conf['name']
        self.start_cash = float(self.conf['start_cash'])

        if self.agent is None:
            self.set_agent(SimulatedAgent(self.market, None), OrderCost())

        # for backtrade, we force to use simulated agent
        if (not realtime_trade) and (self.agent.agent_type != 'simulated'):
            self.set_agent(SimulatedAgent(self.market, None), self.agent.order_cost)

        self.agent.init_portfolio( self.start_cash )

        self.max_value = self.agent.get_portfolio().total_value()

        columns = ['buy', 'sell', 'value', 'cash', 'drawdown']
        self.stat_df = pd.DataFrame([], columns = columns, index = pd.to_datetime([]))
        self.update_stat(self.market.current_date - dt.timedelta(days=1))

        self.load_strategy( self.conf['strategy'] )

    def update_stat(self, date = None):
        # calculate current value of today, save to dataframe
        agent = self.agent
        portfolio = agent.get_portfolio()
        value = portfolio.total_value()
        self.max_value = max(value, self.max_value)
        buy_count, sell_count = agent.get_transaction_count()
        row = [buy_count, - sell_count, value, portfolio.available_cash, (value / self.max_value -1)]

        if date is None:
            date = self.market.current_date

        self.stat_df.loc[ pd.to_datetime(date) ] = pd.Series(row, self.stat_df.columns)

    def before_day(self):
        self.agent.before_day()

    def before_tick(self):
        self.agent.before_tick()

    def after_tick(self):
        self.agent.after_tick()

    def after_day(self):
        self.agent.after_day()
        self.update_stat()

    def get_summary(self):
        df = self.stat_df
        if self.verbose:
            print(df)

        start_value = df.value.iloc[0]
        current_value = df.value.iloc[-1]
        earn = current_value / start_value -1

        days = (df.index[-1] - df.index[0]).days
        years = days / 365
        annual_earn = earn_to_annual(earn, days)

        return {
            'fund_id': self.fund_id,
            'fund_name': self.fund_name,
            'strategy': self.strategy.name,
            'start value': '{:,}'.format(round(start_value, 2)),
            'current value': '{:,}'.format(round(current_value, 2)),
            'invest year': round(years, 1),
            'trade count': int(sum(df.buy) - sum(df.sell)),
            'total earn': str(round(100 * earn, 2)) + ' %',
            'annual earn': str(round(100 * annual_earn, 2)) + ' %',
            'max drawdown': str(round(100 * min(df.drawdown), 2)) + ' %',
        }

    def get_stat(self):
        return self.stat_df



# -*- coding: utf-8; py-indent-offset:4 -*-
import os
import sys
from tabulate import tabulate
from ..core.data_cache import *

def cli_finance(params, options):
    syntax_tips = '''Syntax:
    __argv0__ finance <action> <symbols | stockpool.csv | all> [options]

Action:
    update ......................... download finance reports and update finance indicators
    view ........................... view finance indicators with filters

Symbols:
    <symbols> ...................... stock symbol list, like 600036 000002
    <stockpool.csv> ................ stock pool csv file
    all ............................ all symbols

Options:
    -sortby=<col> .................. sort by the column

    -ipo_years=<from>-<to> ......... IPO years between <from> to <to>
    -earn_ttm=<from>-<to> .......... annual earn between <from> to <to>
    -roe=<from>-<to> ............... ROE between <from> to <to>
    -grow_rate=<from>-<to> ......... average yearly grow rate between <from> to <to>
    -3yr_grow_rate=<from>-<to> ..... recent 3 year grow rate between <from> to <to>

    -tab ........................... show data in table format

    -out=<out.csv> ................. export selected stocks into <out.csv> file

Example:
    __argv0__ finance show 600036 000002 600276
    __argv0__ finance show my-stocks.csv
    __argv0__ finance show all -ipo_years=1- -earn_ttm=1.0- -roe=0.15- -3yr_grow_rate=0.15- -sortby=roe -out=good_stock.csv
'''.replace('__argv0__',os.path.basename(sys.argv[0]))

    if (len(params) == 0) or (params[0] == 'help'):
        print(syntax_tips)
        return

    action = params[0]
    params = params[1:]

    if action not in ['update', 'view', 'show', 'filter']:
        print('\nError: invalid action: ', action)
        return

    if params[0] == 'all':
        df = get_finance_indicator_all(force_update= (action == 'update'))
    else:
        if params[0].endswith('.csv'):
            stock_df = pd.read_csv(params[0], dtype=str)
            symbols = stock_df['symbol'].tolist()
        else:
            symbols = params
        df = get_finance_indicator_df(symbols, force_update= (action == 'update'))

    total_n = df.shape[0]

    # now filter
    df = filter_with_options(df, options)
    filtered_n = df.shape[0]

    # now sort
    df = sort_with_options(df, options, by_default='roe')

    if '-tab' in options:
        if filtered_n > 30:
            print( tabulate(df.head(15), headers='keys', tablefmt='psql') )
            print( ' ...... too many data to show ......')
            print( tabulate(df.tail(15), headers='keys', tablefmt='psql') )
        else:
            print( tabulate(df, headers='keys', tablefmt='psql') )
    else:
        print('-' * 80)
        print(df)
        print('-' * 80)


    print('{} out of {} records selected.'.format(filtered_n, total_n))

    out_csv_file = ''
    for k in options:
        if k.startswith('-out=') and k.endswith('.csv'):
            out_csv_file = k.replace('-out=', '')
    if out_csv_file:
        df = df[['symbol', 'name']]
        df.to_csv(out_csv_file, index= False)
        print('Exported to:', out_csv_file)
        print(df)

    print('')



# -*- coding: utf-8 -*-

import pandas as pd

def signal_to_long(signal, long_value = 1, short_value = 0):
    trend = [0 for i in range(signal.size)]
    for i in range(0, signal.size):
        s = signal.iloc[i]
        if s > 0:
            trend[i] = long_value
        elif s < 0:
            trend[i] = short_value
        elif i > 0:
            trend[i] = trend[i-1]
    return pd.Series(trend, index=signal.index)

def long_to_signal(trend):
    return trend.diff(1).fillna(0)

_registered_signal_indicators = {
    # id: {
    # 'func': function(),
    # 'cols': [],
    # 'type': '',
    # 'label': '',
    # }
}

def get_all_signal_indicators():
    return _registered_signal_indicators

def list_signal_indicators():
    return _registered_signal_indicators.keys()

def register_signal_indicator(id, func, cols, label, ind_type):
    _registered_signal_indicators[ id ] = {
        'func': func,
        'cols': cols,
        'type': ind_type,
        'label': label,
    }
 
def gen_indicator_signal(df, indicators, inplace=False):
    signal = pd.Series(0, index=df.index)
    for k in indicators:
        if k in _registered_signal_indicators:
            values = _registered_signal_indicators[k]
            signal_func = values['func']
            # use + to merge the signals
            signal += signal_func(df, inplace=inplace)
            if len(indicators) == 1:
                return signal
        else:
            print(k, 'not found in registered indicators')

    long = signal_to_long(signal)
    return long.diff(1).fillna(0)



# -*- coding: utf-8; py-indent-offset:4 -*-

import os

from ..utils import *
from .agent_simulated import *

class HumanAgent(SimulatedAgent):
    market = None
    conf = None

    last_load_modified_time = None
    last_order_time = None

    order_file = ''
    transaction_file = ''
    human_position_file = ''
    bot_position_file = ''

    push_service = None

    def set_verbose(self, verbose = True):
        super().set_verbose(verbose)
        if self.push_service is not None:
            self.push_service.set_verbose(verbose)

    def __init__(self, market, agent_conf = None):
        super().__init__(market, agent_conf)

        self.agent_type = 'simulated'

        self.account = self.conf['account']
        self.order_file = self.conf['order'].replace('{account}', self.account)
        self.portfolio_load_file = self.conf['portfolio_load'].replace('{account}', self.account)
        self.portfolio_save_file = self.conf['portfolio_save'].replace('{account}', self.account)

    def init_portfolio(self, start_cash):
        self.portfolio = Portfolio(self.market)
        self.last_order_time = self.market.current_time
        if os.path.isfile(self.portfolio_load_file):
            load_modified_time = get_file_modify_time(self.portfolio_load_file)
            self.portfolio.from_csv( self.portfolio_load_file )
            self.last_load_modified_time = load_modified_time
            print('\n... {} |'.format(str_now()), 'loaded:', self.portfolio_load_file, load_modified_time)
        else:
            self.portfolio.available_cash = start_cash

    def before_day(self):
        super().before_day()

    def after_day(self):
        super().after_day()

    def before_tick(self):
        super().before_tick()

    def after_tick(self):
        super().after_tick()

        if os.path.isfile(self.portfolio_save_file):
            save_modified_time = get_file_modify_time(self.portfolio_save_file)
            need_save = (self.last_order_time > save_modified_time)
        else:
            need_save = True

        if need_save:
            self.portfolio.to_csv( self.portfolio_save_file )
            print('\n... {} |'.format(str_now()), 'updated:', self.portfolio_save_file)

        if os.path.isfile(self.portfolio_load_file):
            load_modified_time = get_file_modify_time(self.portfolio_load_file)
            need_load = (load_modified_time > self.last_load_modified_time)
            if need_load:
                self.portfolio = Portfolio(self.fund)
                self.portfolio.from_csv( self.portfolio_load_file )
                self.last_load_modified_time = load_modified_time
                print('\n... {} |'.format(str_now()), 'reloaded:', self.portfolio_load_file, load_modified_time)

    def exec_order(self, symbol, real_count, real_price, earn_str = '', comment = ''):
        market = self.market
        name = market.get_name(symbol)

        self.last_order_time = market.current_time

        str_now = market.current_time.strftime('%Y-%m-%d %H:%M:%S')
        trade = LANG('buy') if real_count > 0 else LANG('sell')
        msg = '{} | {} {} | {}: {} * {} | {} {}'.format(str_now, symbol, name, trade, round(real_count), round(real_price,2), earn_str, comment)
        print('\r'+msg)

        if self.order_file:
            fp = open(self.order_file, 'a+')
            fp.write(msg + '\n')
            fp.close()

        if self.push_service is not None:
            self.push_service.add_order(symbol, name, real_count, real_price, earn_str, comment)




import os
import sys
from ..core.data_cache import *

def init_hiquant_conf(config_file):
    hiquant_conf_template = '''
[main]
lang = zh
market = cn, hk

[order_cost]
close_tax = 0.001
open_commission = 0.0003
close_commission = 0.0003
min_commission = 5

[indicator]
# add your own indicators here
# my_abc = indicator/my_abc.py

[strategy]
# add your own strategy here
# 001 = strategy/001_macd.py
'''

    print('  Creating', config_file)
    fp = open(config_file, 'w')
    fp.write(hiquant_conf_template)
    fp.close()
    pass

def cli_create(params, options):
    syntax_tips = '''Syntax:
    __argv0__ create <folder>

Example:
    __argv0__ create myProj
'''.replace('__argv0__',os.path.basename(sys.argv[0]))

    if (len(params) == 0) or (params[0] == 'help'):
        print(syntax_tips)
        return

    folder = params[0]
    if os.path.exists(folder):
        print('Folder', folder, 'already exists.\n')
        return

    # create following folder structure
    '''
hiquant.conf
cache/
cache/finance/
cache/market/
cache/pepb/
cache/bank/
cache/A_index_all.csv
cache/A_stock_all.csv
etc/
stockpool/
strategy/
data/
output/
log/
'''

    # create working folder first
    print('Initializing hiquant project folder ...')
    print('  Creating', folder)
    os.mkdir(folder)

    # change to current working folder
    os.chdir(folder)

    # init hiquant config file
    init_hiquant_conf( 'hiquant.conf' )

    # create sub folders
    for sub in [
        'cache',
        'cache/finance',
        'cache/market',
        'cache/pepb',
        'cache/bank',
        'etc',
        'stockpool',
        'strategy',
        'data',
        'output',
        'log',
        ]:
        subfolder = sub
        print('  Creating', subfolder)
        os.mkdir(subfolder)

    # download all stock list and index list
    print('Downloading China A-market stock list and index list ...')
    get_all_stock_list_df()
    get_all_index_list_df()

    print('Done.\n')



# -*- coding: utf-8 -*-

from .basic import CROSS, NON_ZERO, MAD
from .ma import SMA
from ..core.indicator_signal import register_signal_indicator

# Commodity Channel Index, by Donald Lambert, 1980s
# --------------------------------------------
def CCI(close, high, low, length = 14, c = None):
    c = float(c) if c and c > 0 else 0.015
    tp = (high + low + close) / 3.0
    mtp = SMA(tp, length)
    mad_tp = MAD(tp, length)
    cci = (tp - mtp) / NON_ZERO(mad_tp * c)
    return cci

def signal_cci(df, inplace=False, length=14):
    cci = CCI(df.close, df.high, df.low, length=length)
    if inplace:
        df['cci'], df['cci_upper'], df['cci_lower'] = cci, 100, -100
    cross_up = CROSS(cci, 100)
    cross_md = CROSS(cci, 0)
    cross_dn = CROSS(cci, -100)
    buy_signal = (cross_up > 0) | (cross_md > 0) | (cross_dn > 0)
    sell_signal = (cross_up < 0) | (cross_md < 0) | (cross_dn < 0)
    return buy_signal.astype(int) - sell_signal.astype(int)

register_signal_indicator('cci', signal_cci, ['cci', 'cci_upper', 'cci_lower'], 'CCI', 'obos')



# -*- coding: utf-8 -*-

import pandas as pd
import numpy as np
from sys import float_info as sflt

# --------------------------------------------
def MAX(data, length):
    return data.rolling(length, min_periods=1).max()

# --------------------------------------------
def MIN(data, length):
    return data.rolling(length, min_periods=1).min()

# --------------------------------------------
def REF(data, length):
    return data.shift(length).fillna(method='bfill')

# --------------------------------------------
def NON_ZERO(data):
    if data.eq(0).any():
        return data + sflt.epsilon
    return data

# --------------------------------------------
def SIGN(data):
    data = data.copy()
    data[data > 0] = 1
    data[data < 0] = -1
    return data

# --------------------------------------------
def DIFF_SIGN(data, initial = 0, drift = 1):
    sign = data.diff(drift)
    sign[sign > 0] = 1
    sign[sign < 0] = -1
    sign.iloc[0] = initial
    return sign

# --------------------------------------------
def SUM(data, length):
    return data.rolling(length, min_periods=1).sum()

# ---------------------------------------------------------
def ADD(data1, data2):
    return data1 + data2

# ---------------------------------------------------------
def SUB(data1, data2):
    return data1 - data2

# ---------------------------------------------------------
def MUL(data1, data2):
    return data1 * data2

# ---------------------------------------------------------
def DIV(data1, data2):
    return data1 / NON_ZERO(data2)

# --------------------------------------------
def ABS(data):
    return data.abs()

# --------------------------------------------
def POW(data, x):
    return data.pow(x)

# --------------------------------------------
def VAR(data, length, ddof=None):
    ddof = int(ddof) if ddof and ddof >= 0 and ddof < length else 0
    var = data.rolling(length, min_periods=1).var(ddof)
    return var

# --------------------------------------------
def STDEV(data, length, ddof=None):
    ddof = int(ddof) if ddof and ddof >= 0 and ddof < length else 0
    stdev = VAR(data, length, ddof=ddof).apply(np.sqrt)
    return stdev

# --------------------------------------------
def MAD(data, length, ddof=None):
    def _mad(series):
        return np.fabs(series - series.mean()).mean()

    mad = data.rolling(length, min_periods=1).apply(_mad, raw=True)
    return mad

# return, up = 1, down = -1, no = 0
# --------------------------------------------
def CROSS(fast, slow):
    d = fast - slow
    d[ d <= 0 ] = 0
    d[ d > 0 ] = 1
    return d.diff(1)

# return, top = 1, bottom = -1, no = 0
# --------------------------------------------
def PEAK(data):
    # we cannot predict future, so shift 2 days
    data = REF(data, 2)
    data = DIFF_SIGN(data,drift=1) + DIFF_SIGN(data,drift=-1) + DIFF_SIGN(data,drift=2) + DIFF_SIGN(data,drift=-2)
    # if >= 3, large than near 3 or 4 values, we regard it a peak
    return data // 3




# -*- coding: utf-8; py-indent-offset:4 -*-

import os
import datetime as dt

def str_now():
    return dt.datetime.now().strftime('%Y:%m:%d %H:%M:%S')

def datetime_today():
    now = dt.datetime.now()
    return dt.datetime(now.year, now.month, now.day)

def get_file_modify_time(fname):
    ts = os.stat(fname).st_mtime
    return dt.datetime.fromtimestamp(ts)

# 20180101
# 1 year ago
# 2 days later
# future
# etc.
def date_from_str(str):
    today = datetime_today()
    if ('ago' in str) or ('later' in str):
        items = str.split(' ')
        n = int(items[0])
        unit = items[1]
        ago_later = +1 if (items[2] == 'later') else -1
        if 'day' in items[1]:
            return today + ago_later * dt.timedelta(days=n)
        elif 'month' in items[1]:
            return today + ago_later * dt.timedelta(days=n*30)
        elif 'year' in items[1]:
            return today + ago_later * dt.timedelta(days=n*365)
        elif 'week' in items[1]:
            return today + ago_later * dt.timedelta(days=n*7)
        else:
            raise ValueError('invalid format:' + str)
    elif 'today' == str:
        return today
    elif 'future' == str:
        return today + dt.timedelta(days=99*365)
    elif 'yesterday' == str:
        return today - dt.timedelta(days=1)
    elif 'tomorrow' == str:
        return today + dt.timedelta(days=1)
    else:
        return dt.datetime.strptime(str, '%Y%m%d')

# 00:01:00
# 1 hour/day/week/month/year
# 5 min
# 30 sec
def seconds_from_str(str):
    if ':' in str:
        d = dt.datetime.strptime(str, '%Y:%M:%S')
        return (d - dt.datetime(d.year, d.month, d.day)).total_seconds()
    else:
        items = str.split(' ')
        n = int(items[0])
        unit = items[1]
        if 'sec' in unit:
            return n * 1
        elif 'min' in unit:
            return n * 60
        elif 'hour' in unit:
            return n * 3600
        elif 'day' in unit:
            return n * 3600 * 24
        elif 'week' in unit:
            return n * 3600 * 24 * 7
        elif 'month' in unit:
            return n * 3600 * 24 * 30
        elif 'year' in unit:
            return n * 3600 * 24 * 365
        else:
            raise ValueError('invalid format:' + str)



# -*- coding: utf-8; py-indent-offset:4 -*-

import os
import sys
import datetime as dt
import pandas as pd
import matplotlib.pyplot as plt
import talib as tl
from talib import abstract
from tabulate import tabulate

from ..utils import dict_from_df
from ..core import get_all_index_list_df, get_all_stock_list_df
from ..indicator import SIGN

_talib_pattern_length = {
'CDL2CROWS': 3,
'CDL3BLACKCROWS': 3,
'CDL3INSIDE': 3,
'CDL3LINESTRIKE': 4,
'CDL3OUTSIDE': 3,
'CDL3STARSINSOUTH': 3,
'CDL3WHITESOLDIERS': 3,
'CDLABANDONEDBABY': 3,
'CDLADVANCEBLOCK': 3,
'CDLBELTHOLD': 2,
'CDLBREAKAWAY': 5,
'CDLCLOSINGMARUBOZU': 1,
'CDLCONCEALBABYSWALL': 4,
'CDLCOUNTERATTACK': 2,
'CDLDARKCLOUDCOVER': 2,
'CDLDOJI': 1,
'CDLDOJISTAR': 1,
'CDLDRAGONFLYDOJI': 1,
'CDLENGULFING': 2,
'CDLEVENINGDOJISTAR': 3,
'CDLEVENINGSTAR': 3,
'CDLGAPSIDESIDEWHITE': 2,
'CDLGRAVESTONEDOJI': 1,
'CDLHAMMER': 1,
'CDLHANGINGMAN': 1,
'CDLHARAMI': 2,
'CDLHARAMICROSS': 2,
'CDLHIGHWAVE': 3,
'CDLHIKKAKE': 3,
'CDLHIKKAKEMOD': 3,
'CDLHOMINGPIGEON': 2,
'CDLIDENTICAL3CROWS': 3,
'CDLINNECK': 2,
'CDLINVERTEDHAMMER': 1,
'CDLKICKING': 2,
'CDLKICKINGBYLENGTH': 2,
'CDLLADDERBOTTOM': 5,
'CDLLONGLEGGEDDOJI': 1,
'CDLLONGLINE': 1,
'CDLMARUBOZU': 1,
'CDLMATCHINGLOW': 2,
'CDLMATHOLD': 5,
'CDLMORNINGDOJISTAR': 3,
'CDLMORNINGSTAR': 3,
'CDLONNECK': 2,
'CDLPIERCING': 2,
'CDLRICKSHAWMAN': 1,
'CDLRISEFALL3METHODS': 5,
'CDLSEPARATINGLINES': 2,
'CDLSHOOTINGSTAR': 1,
'CDLSHORTLINE': 1,
'CDLSPINNINGTOP': 1,
'CDLSTALLEDPATTERN': 3,
'CDLSTICKSANDWICH': 3,
'CDLTAKURI': 1,
'CDLTASUKIGAP': 3,
'CDLTHRUSTING': 2,
'CDLTRISTAR': 3,
'CDLUNIQUE3RIVER': 3,
'CDLUPSIDEGAP2CROWS': 3,
'CDLXSIDEGAP3METHODS': 5,
}

def list_existing_daily_files():
    file_list = []
    files = os.listdir('cache/market')
    for file in files:
        if file.endswith('_daily.csv'):
            file_list.append('cache/market/'+file)
    return file_list

def cli_pattern_list():
    all_patterns = tl.get_function_groups()['Pattern Recognition']
    table = []
    for pattern in all_patterns:
        info = _talib_pattern_length[ pattern ]
        func = abstract.Function(pattern)
        table.append([pattern, func.info['display_name'], info])
    df = pd.DataFrame(table, columns=['pattern', 'name', 'length'])
    print( tabulate(df, headers='keys', tablefmt='psql') )

def cli_pattern_stat():
    all_patterns = tl.get_function_groups()['Pattern Recognition']

    # init the counters
    pattern_counters = {}
    for pattern in all_patterns:
        pattern_counters[ pattern ] = [0, 0]

    # find all daily files
    files = list_existing_daily_files()
    n = len(files)
    i = 0
    for file in files:
        i += 1
        print('\r ... {}/{} ... searching patterns in {} ...'.format(i, n, file), end= '', flush= True)
        df = pd.read_csv(file, parse_dates=['date']).set_index('date', drop=True)
        df = df.astype(float)

        # prepare for validity exam
        diff = df['close'].diff().shift(-1).fillna(0)

        for pattern in all_patterns:
            # find the pattern signals
            func = abstract.Function(pattern)
            signal = func(df)

            # examine
            verify = SIGN( signal * diff )
            valid_signals = verify[ verify > 0 ].sum()
            invalid_signals = verify[ verify < 0 ].abs().sum()

            pattern_counters[ pattern ][0] += valid_signals
            pattern_counters[ pattern ][1] += invalid_signals
    print('\n')

    # stat result as dataframe
    table = []
    for pattern, counters in pattern_counters.items():
        sum_counter = counters[0] + counters[1]
        valid_rate = ' - ' if (sum_counter == 0) else (str(round(counters[0] / sum_counter * 100.0)).rjust(3) + " %")
        row = [pattern, counters[0], counters[1], valid_rate]
        table.append(row)
    stat_df = pd.DataFrame(table, columns=['pattern', 'valid', 'invalid', 'valid_rate'])

    stat_df.sort_values('valid_rate', ascending=False, inplace=True)

    print( tabulate(stat_df, headers='keys', tablefmt='psql') )

def find_pattern(func):
    files = list_existing_daily_files()
    for file in files:
        df = pd.read_csv(file, parse_dates=['date']).set_index('date', drop=True)
        df = df.astype(float)

        signal = func(df)
        signal = signal[ signal != 0 ]
        if len(signal) > 0:
            symbol = os.path.basename(file).replace('_daily.csv','')
            return symbol, signal, df

    return None, [], None

def get_k_value(price, pattern_date, length):
    '''将开、高、低、收价格数据进行处理，以支持K线图展示
    Args:
        pattern: 形态函数返回的标的与时间元组
        count: 形态数量
    Return:
        返回处理后的价格数据，类型为df
    '''

    start = pattern_date - dt.timedelta(days=20)
    end = pattern_date + dt.timedelta(days=20)
    price = price[start:end]

    dates = price.index.tolist()
    i = dates.index(pattern_date)
    pattern_dates = dates[i+1-length:i+1]

    y_list = {'y1': [], 'y2': [], 'y3': [], 'y4': [], 'c': [], 'al': []}
    
    # 数据处理
    for i in range(price.shape[0]):
        # 计算各bar的透明度
        #print(price.index[i].date(), type(price.index[i].date()))
        if price.index[i] in pattern_dates:
            y_list['al'].append(1)
        else:
            y_list['al'].append(0.3)
        
        # 计算各实体的长度
        if price.open[i] > price.close[i]:
            y_list['y1'].append(price.close[i])
            y_list['y2'].append(price.open[i] - price.close[i])
            y_list['c'].append('g')
        else:
            y_list['y1'].append(price.open[i])
            y_list['y2'].append(price.close[i] - price.open[i])
            y_list['c'].append('r')
        
        # 计算各影线的长度
        if price.high[i] > price.low[i]:
            y_list['y3'].append(price.low[i])
            y_list['y4'].append(price.high[i] - price.low[i])
        else:
            y_list['y3'].append(price.high[i])
            y_list['y4'].append(price.low[i] - price.high[i])
            
    for name, value in y_list.items():
        price.loc[:, name] = value

    return price

def show_bar(data, title='None', out_file = None):
    '''将价格数据按实体与影线进行组合
    Args:
        data: 处理好的价格数据，df类型
        title: 图片标题，可选
    Return:
        None
    '''

    # 整合 x 轴与 y 轴数据
    x = [str(date.date()) for date in data.index]
    y1 = data['y1']
    y2 = data['y2']

    y3 = data['y3']
    y4 = data['y4']
    
    fig = plt.figure(figsize=(10, 5))

    # 画影线，要求对应不同的颜色与透明度
    for i in range(len(data)):
        plt.bar(x[i], y3[i], align='center', alpha=0)
        plt.bar(x[i], y4[i], align='center', color=data['c'][i], 
                alpha=data['al'][i], bottom=y3[i], width=0.1)

    # 画实体，要求对应不同的颜色与透明度
    for j in range(len(data)):
        plt.bar(x[j], y1[j], align='center', alpha=0)
        plt.bar(x[j], y2[j], align='center', color=data['c'][j], 
                alpha=data['al'][j], bottom=y1[j])
    
    _min = data[['open', 'high', 'low', 'close']].min().min() - 1.0
    _max = data[['open', 'high', 'low', 'close']].max().max() + 1.0

    plt.ylim(_min, _max)
    plt.margins(x=0.0, y=0.05)
    plt.title(title)
    plt.xticks(rotation=45)

    if out_file is None:
        plt.show()
    else:
        plt.savefig(out_file)

# 检查某标的在某日形成的形态，返回形态描述
def discern_pattern(daily_df, date):
    price = daily_df[ daily_df.index <= date ]
    pattern_function = tl.get_function_groups()['Pattern Recognition']
    for name in pattern_function:
        func = abstract.Function(name)
        signal_series = func(price)
        if signal_series[-1] != 0:
            yield func.info['display_name']

def cli_pattern_plot(pattern, out_file = None):
    all_patterns = tl.get_function_groups()['Pattern Recognition']
    pattern = all_patterns[int(pattern) % len(all_patterns)] if pattern.isdigit() else pattern
    if pattern not in all_patterns:
        print('\nError: pattern "{}" not exists'.format(pattern))
        return False

    func = abstract.Function(pattern)
    symbol, signal, df = find_pattern(func)

    if len(signal) > 0:
        # symbol -> name mapping of all stocks and index
        stock_df = get_all_stock_list_df()
        index_df = get_all_index_list_df()
        stock_df = stock_df.append(index_df, ignore_index= True)
        symbol_name = dict_from_df(stock_df, 'symbol', 'name')

        name = symbol_name[symbol] if (symbol in symbol_name) else ''

        # plot the pattern
        length = _talib_pattern_length[pattern]
        k_value = get_k_value(df, signal.index[0], length)
        title = '{} - {} : {}'.format(symbol, name, func.info['display_name'])
        show_bar(k_value, title, out_file= out_file)
        return True

    else:
        print('\nPattern "{}" not found after searching all daily data in cache/market .'.format(pattern))
        print('You can try again after download more daily data.\n')
        return False

def cli_pattern(params, options):
    syntax_tips = '''Syntax:
    __argv0__ pattern list
    __argv0__ pattern stat
    __argv0__ pattern demo <pattern>

Actions:
    list ........................... list pattern supported in this tool
    stat ........................... count all the patterns and effectiveness
    demo ........................... plot pattern as demo

Pattern:
    i .............................. index of patterns, range 0 - 60
    name ........................... name of pattern, like CDL2CROWS

Example:
    __argv0__ pattern list
    __argv0__ pattern demo 1
'''.replace('__argv0__',os.path.basename(sys.argv[0]))

    if (len(params) == 0) or (params[0] == 'help'):
        print( syntax_tips )
        return

    action = params[0]
    params = params[1:]

    if action == 'list':
        cli_pattern_list()

    elif action == 'stat':
        cli_pattern_stat()

    elif action in ['demo', 'plot']:
        out_file = None
        for option in options:
            if option.startswith('-out=') and option.endswith('.png'):
                out_file = option.replace('-out=', '')

        if len(params) > 0:
            if params[0] == 'all':
                all_patterns = tl.get_function_groups()['Pattern Recognition']
                for pattern in all_patterns:
                    out_file = 'output/{}.png'.format(pattern)
                    if cli_pattern_plot(pattern, out_file= out_file):
                        print('pattern "{}" saved to: {}'.format(pattern, out_file))
            else:
                cli_pattern_plot(params[0], out_file= out_file)
        else:
            print('\nError: pattern index or name exptected.')

    else:
        print('\nError: invalid action:', action)



# -*- coding: utf-8; py-indent-offset:4 -*-

import os
from ..utils import datetime_today
from .lang import LANG

class Strategy:
    fund = None
    name = ''
    conf = {}

    max_stocks = 10
    max_weight = 1.2
    stop_loss = 0.0
    stop_earn = 10000.0

    def __init__(self, fund, strategy_file):
        fund.strategy = self

        self.fund = fund
        self.name = os.path.basename(strategy_file)
        self.conf = fund.conf

        self.max_stocks = 10
        self.max_weight = 1.0
        self.stop_loss = 1 + (-1.0)
        self.stop_earn = 1 + (10000.0)

        if 'max_stocks' in self.conf:
            self.max_stocks = int(self.conf['max_stocks'])
        if 'max_weight' in self.conf:
            self.max_weight = float(self.conf['max_weight'])
        if 'stop_loss' in self.conf:
            self.stop_loss = 1 + float(self.conf['stop_loss'])
        if 'stop_earn' in self.conf:
            self.stop_earn = 1 + float(self.conf['stop_earn'])

class BasicStrategy( Strategy ):
    symbol_signal = {}

    def __init__(self, fund, strategy_file):
        super().__init__(fund, strategy_file)

        self.targets = self.select_stock()
        self.symbol_signal = {}

        market = fund.market
        agent = fund.agent
        portfolio = agent.get_portfolio()

        concerned_stocks = list(set(self.targets) | set(portfolio.positions.keys()))
        concerned_stocks.sort()
        market.watch( concerned_stocks )

    def select_stock(self):
        pass

    def gen_trade_signal(self, symbol, init_data = False):
        pass

    def get_signal_comment(self, symbol, signal):
        pass

    def trade(self):
        fund = self.fund
        market = fund.market
        agent = fund.agent
        portfolio = agent.get_portfolio()

        total_value = portfolio.total_value()
        max_stocks = min(self.max_stocks, len(self.targets))
        max_pos_per_stock = total_value / max_stocks * self.max_weight

        # calc and cache the trade signal for history data
        concerned_stocks = list(set(self.targets) | set(portfolio.positions.keys()))
        concerned_stocks.sort()
        market.watch( concerned_stocks )
        for symbol in concerned_stocks:
            if not (symbol in self.symbol_signal):
                if self.fund.verbose:
                    print('\tinit history trade signal:', symbol)
                self.symbol_signal[ symbol ] = self.gen_trade_signal(symbol, init_data=True)

        # find out the stocks to sell or buy
        sell_list = {}
        buy_list = {}
        for symbol in concerned_stocks:
            if not market.allow_trading(symbol, market.current_date):
                continue

            price = market.get_price(symbol, market.current_date)
            if price < 0.01:
                continue

            if market.current_date < datetime_today():
                trade_signal = self.symbol_signal[ symbol ].loc[ market.current_date ]
            else:
                trade_signal = self.gen_trade_signal(symbol, init_data=False).iloc[-1]

            if (symbol in portfolio.positions):
                cost = portfolio.positions[ symbol ].cost
                if price <= cost * self.stop_loss:
                    diff = price / cost - 1
                    sell_list[ symbol ] = LANG('stop loss') + ': {} %'.format(round(100*diff,2))
                elif price >= cost * self.stop_earn:
                    diff = price / cost - 1
                    sell_list[ symbol ] = LANG('stop earn') + ' {} %'.format(round(100*diff,2))
                elif (trade_signal < 0):
                    sell_list[ symbol ] = LANG('signal') + self.get_signal_comment(symbol, trade_signal)

            if (trade_signal > 0) and (symbol in self.targets) and (not symbol in sell_list):
                stock_value = 0.0
                if symbol in portfolio.positions:
                    stock_value = price * portfolio.positions[ symbol ].shares
                if stock_value < max_pos_per_stock:
                    buy_list[ symbol ] = LANG('signal') + self.get_signal_comment(symbol, trade_signal)

        # now sell first, we may need the cash to buy other stocks
        for symbol, reason in sell_list.items():
            agent.order_target(symbol, 0, comment = reason)

        # now buy
        for symbol, reason in buy_list.items():
            agent.order_target_value(symbol, max_pos_per_stock, comment = reason)



# -*- coding: utf-8; py-indent-offset:4 -*-

import os
import sys
from tabulate import tabulate
from ..core.data_cache import *

def cli_pepb(params, options):
    syntax_tips = '''Syntax:
    __argv0__ pepb <action> <symbols | symbos.csv | all> [options]

Action:
    update ................................. download PE/PB data and update PE/PE calculation
    view ................................... view PE/PB data with filters

Symbols:
    <symbols> .............................. stock symbol list, like 600036 000002
    <symbols.csv> .......................... stock symbol list csv file
    all .................................... all symbols

Options:
    -sortby=<col> .......................... sort by the column

    -pe=<from>-<to> ........................ sort by PE
    -pb=<from>-<to> ........................ sort by PB
    -pe_pos=<from>-<to> .................... sort by PE % position
    -pb_pos=<from>-<to> .................... sort by PB % position

    -out=<out.csv> ......................... export selected stocks into <out.csv> file

Example:
    __argv0__ pepb show 600036 000002 600276
    __argv0__ pepb show my-stocks.csv -pe_pos
    __argv0__ pepb show all
'''.replace('__argv0__',os.path.basename(sys.argv[0]))

    if (len(params) == 0) or (params[0] == 'help'):
        print( syntax_tips )
        return

    action = params[0]
    params = params[1:]

    if action not in ['update', 'view', 'show', 'filter']:
        print('\nError: invalid action: ', action)
        return

    if params[0] == 'all':
        df = get_pepb_symmary_all(force_update= (action == 'update'))
    else:
        if params[0].endswith('.csv'):
            stock_df = pd.read_csv(params[0], dtype=str)
            symbols = stock_df['symbol'].tolist()
        else:
            symbols = params
        df = get_pepb_symmary_df(symbols, force_update= (action == 'update'))

    total_n = df.shape[0]

    # now filter
    df = filter_with_options(df, options)
    filtered_n = df.shape[0]

    # now sort
    df = sort_with_options(df, options, by_default='pb_pos')

    print( tabulate(df, headers='keys', tablefmt='psql') )

    print('{} out of {} records selected.'.format(filtered_n, total_n))

    out_csv_file = ''
    for k in options:
        if k.startswith('-out=') and k.endswith('.csv'):
            out_csv_file = k.replace('-out=', '')
    if out_csv_file:
        df = df[['symbol', 'name']]
        df.to_csv(out_csv_file, index= False)
        print('Exported to:', out_csv_file)
        print(df)

    print('')



# -*- coding: utf-8 -*-

from .basic import CROSS, NON_ZERO, MIN, MAX
from ..core.indicator_signal import register_signal_indicator

# William %R, by Larry Williams, 1973
# --------------------------------------------
def WR(close, high, low, length = 30):
    lowest_low = MIN(low, length)
    highest_high = MAX(high, length)
    wr = (highest_high - close) / NON_ZERO(highest_high - lowest_low) * 100.0
    return wr

def signal_wr(df, inplace=False, length=10):
    wr = WR(df.close, df.high, df.low, length=length)
    if inplace:
        df['wr'], df['wr_upper'], df['wr_lower'] = wr, 80, 20
    buy_signal = CROSS(wr, 80) > 0
    sell_signal = CROSS(wr, 20) < 0
    return buy_signal.astype(int) - sell_signal.astype(int)

register_signal_indicator('wr', signal_wr, ['wr', 'wr_upper', 'wr_lower'], 'W%R', 'obos')



# -*- coding: utf-8; py-indent-offset:4 -*-

from ..utils import *
from .stock import *
from .order_cost import *
from .portfolio import *
from .push_master import *

class SimulatedAgent:
    conf = {}
    agent_type = ''

    market = None
    order_cost = None
    push_service = None

    portfolio = None
    sell_count = 0
    buy_count = 0

    verbose = False

    def set_verbose(self, verbose = True):
        self.verbose = verbose

    def set_market(self, market):
        self.market = market

    def add_push_service(self, push_servcie):
        self.push_service.add_service(push_servcie)

    def set_order_cost(self, order_cost):
        self.order_cost = order_cost

    def __init__(self, market, agent_conf = None):
        self.agent_type = 'simulated'
        self.market = market
        self.conf = agent_conf

        self.portfolio = Portfolio(market)
        self.order_cost = OrderCost(0.001, 0.0003, 0.0003, 5.0)
        self.push_service = MasterPush()

        self.sell_count = 0
        self.buy_count = 0

    def before_day(self):
        self.buy_count = 0
        self.sell_count = 0
        pass

    def after_day(self):
        pass

    def before_tick(self):
        pass

    def after_tick(self):
        if self.market.current_date < datetime_today():
            self.push_service.clear()
        else:
            self.push_service.flush()

    def get_portfolio(self):
        return self.portfolio

    def get_transaction_count(self):
        return self.buy_count, self.sell_count

    def init_portfolio(self, start_cash):
        self.portfolio = Portfolio(self.market)
        self.portfolio.available_cash = start_cash

    def transfer_cash(self, amount, bank_to_security = True):
        if bank_to_security:
            self.portfolio.available_cash += amount
        else:
            self.portfolio.available_cash -= amount
        if self.portfolio.available_cash < 0:
            self.portfolio.available_cash = 0.0

    def exec_order(self, symbol, real_count, real_price, earn_str = '', comment = ''):
        market = self.market
        name = market.get_name(symbol)

        str_now = market.current_time.strftime('%Y-%m-%d %H:%M:%S')
        trade = 'buy' if real_count > 0 else 'sell'
        msg = '{} | {} {} | {}: {} * {} | {} {}'.format(str_now, symbol, name, trade, round(real_count), round(real_price,2), earn_str, comment)
        if self.verbose:
            print('\r'+msg)

    def order(self, symbol, count, price = None, comment = ''):
        market = self.market
        if not price:
            price = market.get_price(symbol)
            if price <= 0:
                return

        p = self.portfolio
        if count > 0 and p.available_cash < price * count:
            # no enough cash to buy stock, ignore order
            return

        real_price = market.get_real_price(symbol)
        real_count = price * count / real_price

        cost = 0
        if symbol in p.positions:
            stock = p.positions[symbol]
            target_count = stock.shares + count
            if target_count > 0:
                # buy, or not sell all, a hand must be times of 100
                real_count = int(real_count / 100) * 100
                if real_count == 0:
                    return

                count = real_count * real_price / price

                stock.cost = (stock.cost * stock.shares + price * count) / target_count
                stock.shares += count
            else:
                # sell all
                count = - stock.shares
                real_count = price * count / real_price

                cost = stock.cost
                stock.cost = 0.0
                stock.shares = 0
                p.history[symbol] = stock
                del p.positions[symbol]
        elif count > 0:
            xcount = count
            # a hand must be times of 100
            real_count = int(real_count / 100) * 100
            if real_count == 0:
                return

            count = real_count * real_price / price

            if symbol in p.history:
                stock = p.history[symbol]
                del p.history[symbol]
            else:
                stock = Stock(symbol, market.get_name(symbol))
            stock.cost = price
            stock.shares = count
            p.positions[symbol] = stock

        else:
            return

        # calculate trade cost
        trade_cost = 0.0
        trade_value = price * max(count, -count)
        if count > 0:
            trade_cost = trade_value * (self.order_cost.open_commission)
        else:
            trade_cost = trade_value * (self.order_cost.close_commission + self.order_cost.close_tax)
        if cost < self.order_cost.min_commission:
            trade_cost = self.order_cost.min_commission

        p.available_cash = round(p.available_cash  - price * count - trade_cost, 2)

        if real_count > 0:
            self.buy_count += 1
        elif real_count < 0:
            self.sell_count += 1

        earn = round(price - cost, 2)
        earn_str = ' --> {} %'.format( round(earn / cost * 100,2) ) if cost > 0 else ''
        self.exec_order(symbol, round(real_count), round(real_price,3), earn_str, comment)

    def order_target(self, symbol, target_count, price = None, comment = ''):
        if target_count < 0:
            return
        if not price:
            market = self.market
            price = market.get_price(symbol)
        if price <= 0:
            return
        p = self.portfolio
        if symbol in p.positions:
            stock = p.positions[symbol]
            count = target_count - stock.shares
        else:
            if target_count == 0:
                return
            count = target_count
        self.order(symbol, count, price, comment)

    def order_value(self, symbol, value, price = None, comment = ''):
        if not price:
            market = self.market
            price = market.get_price(symbol)
        if price <= 0:
            return
        count = value / price
        if count > 0:
            self.order(symbol, count, price, comment)

    def order_target_value(self, symbol, target_value, price = None, comment = ''):
        if not price:
            market = self.market
            price = market.get_price(symbol)
        if price <= 0:
            return
        target_count = target_value / price
        self.order_target(symbol, target_count, price, comment)



# -*- coding: utf-8; py-indent-offset:4 -*-

import os
import sys

def get_strategy_template():
    return '''
import pandas as pd
import talib

from hiquant import *

class MyStrategy( BasicStrategy ):
    def __init__(self, fund, strategy_file):
        super().__init__(fund, strategy_file)
        self.max_stocks = 10
        self.max_weight = 1.2
        self.stop_loss = 1 + (-0.10)
        self.stop_earn = 1 + (+0.20)

    def select_stock(self):
        return ['600036', '300122', '600519', '300357', '601888']

    def gen_trade_signal(self, symbol, init_data = False):
        market = self.fund.market
        if init_data:
            df = market.get_daily(symbol)
        else:
            df = market.get_daily(symbol, end = market.current_date, count = 26+9)

        dif, dea, macd_hist = talib.MACD(df.close, fastperiod=12, slowperiod=26, signalperiod=9)
        return pd.Series( CROSS(dif, dea), index=df.index )

    def get_signal_comment(self, symbol, signal):
        return 'MACD golden cross' if (signal > 0) else 'MACD dead cross'

def init(fund):
    strategy = MyStrategy(fund, __file__)

    trader = fund.trader
    trader.run_daily(before_market_open, strategy, time='before_open')
    trader.run_on_bar_update(trade, strategy)
    trader.run_daily(trade, strategy, time='14:30')
    trader.run_daily(after_market_close, strategy, time='after_close')

def before_market_open(strategy):
    pass

def trade(strategy):
    strategy.trade()

def after_market_close(strategy):
    pass
'''

def cli_strategy_create(params, options):
    template_content = get_strategy_template()

    file_to_create = params[0]
    if os.path.isfile(file_to_create):
        print('\nError: file already exists:', file_to_create)
        return

    fp = open(file_to_create, 'w')
    fp.write(template_content)
    fp.close()

    print('-' * 80)
    print( template_content )
    print('-' * 80)

    print( 'Strategy file created:\n  ', os.path.abspath(file_to_create))
    print( '\nPlease edit file content before running.' )

def cli_strategy_bench(params, options):
    pass

def cli_strategy_clone(params, options):
    pass

def cli_strategy_publish(params, options):
    pass

def cli_strategy(params, options):
    syntax_tips = '''Syntax:
    __argv0__ strategy <action> <my_strategy.py> [options]

Actions:
    create ................ create a strategy python file from template

<my_strategy.py> ............ a strategy python file

Example:
    __argv0__ strategy create strategy/my_strategy.py
    __argv0__ strategy bench strategy/my_strategy.py 20180101
'''.replace('__argv0__',os.path.basename(sys.argv[0]))

    if (len(params) == 0) or (params[0] == 'help'):
        print( syntax_tips )
        return

    action = params[0]
    params = params[1:]

    fund_tools = {
        'create': cli_strategy_create,

        'bench': cli_strategy_bench,
        'clone': cli_strategy_clone,
        'publish': cli_strategy_publish,
    }
    if action in fund_tools.keys():
        if (len(params) == 0) or (not (params[0].endswith('.py'))):
            print('\nError: A strategy filename ending with .py is expected.\n')
            return
        func = fund_tools[ action ]
        func(params, options)
    else:
        print('\nError: unknown action:', action)
        print( syntax_tips )





# -*- coding: utf-8 -*-

from .basic import DIFF_SIGN, NON_ZERO
from .ma import SMA

# On Balance Volume (OBV)
# --------------------------------------------
def OBV(close, volume, length = 30):
    signed_vol = DIFF_SIGN(close, 1) * volume
    obv = signed_vol.cumsum()
    return obv

# Improved OBV
# --------------------------------------------
def OBV2(close, high, low, volume, length = 30):
    obv = ((close - low) - (high - close)) / NON_ZERO(high - low) * volume
    maobv = SMA(obv, length)
    return obv, maobv




import datetime
import pandas as pd
import mplfinance as mpf
import matplotlib as mpl
import matplotlib.pyplot as plt
from cycler import cycler

from ..indicator import *
from .indicator_signal import *

class Stock:
    symbol = ''
    name = ''
    daily_df = None
    shares = 0
    cost = 0.0
    price = 0.0

    def __init__(self, symbol, name = '', daily_df = None):
        self.symbol = symbol
        self.name = name
        self.daily_df = daily_df

    def add_indicator(self, indicators, mix = False, inplace = False, order_cost = None):
        df = self.daily_df
        df['daily_return'] = df.close.pct_change()
        if mix:
            signal = gen_indicator_signal(df, indicators, inplace=inplace)
            long_trend = signal_to_long(signal)
            df['trade_signal'] = signal
            df['long_pos'] = long_trend.shift(1).fillna(0)
            df['act_return'] = df.long_pos * df.daily_return

            # use a column start with "." to store the performance data
            if order_cost is not None:
                sell_cost = signal.copy()
                sell_cost[ sell_cost > 0 ] = 0
                buy_cost = signal.copy()
                buy_cost[ buy_cost < 0 ] = 0
                sell_cost = sell_cost * (order_cost.close_commission + order_cost.close_tax) * (-1)
                buy_cost = buy_cost * order_cost.open_commission
                return_after_cost = df.act_return - (buy_cost + sell_cost)

                df['return.'] = (1 + return_after_cost).cumprod() -1
            else:
                df['return.'] = (1 + df.act_return).cumprod() -1
        else:
            for k in indicators:
                signal = gen_indicator_signal(df, [k], inplace=inplace)
                long_trend = signal_to_long(signal)
                long_pos = long_trend.shift(1).fillna(0)
                act_return = long_pos * df.daily_return

                # use a column start with "." to store the performance data
                if order_cost is not None:
                    sell_cost = signal.copy()
                    sell_cost[ sell_cost > 0 ] = 0
                    buy_cost = signal.copy()
                    buy_cost[ buy_cost < 0 ] = 0
                    sell_cost = sell_cost * (order_cost.close_commission + order_cost.close_tax) * (-1)
                    buy_cost = buy_cost * order_cost.open_commission
                    return_after_cost = act_return - (buy_cost + sell_cost)
                    df[k + '.'] = (1 + return_after_cost).cumprod() - 1
                else:
                    df[k + '.'] = (1 + act_return).cumprod() - 1


    def rank_indicator(self, by = 'overall'):
        df = self.daily_df.copy()

        # fill NaN with 0
        df.fillna(0, inplace=True)

        # sort the indicators by overall performance
        table = []
        for k in df.columns:
            if k.endswith('.'):
                v = df[k]
                final = v.iloc[-1]
                overall = sum(v.tolist())
                table.append([k, final, overall])
        rank_df = pd.DataFrame(table, columns=['indicator', 'final', 'overall'])
        rank_df.sort_values(by=by, ascending=False, inplace=True)
        rank_df.reset_index(inplace=True, drop=True)

        return rank_df

    # Visualization
    # --------------------------------------------------------------------------------
    def plot(self, red_up = True, out_file = None):
        # to correctly display Chinese text
        # download TTF files:
        # https://raw.githubusercontent.com/StellarCN/scp_zh/master/fonts/SimHei.ttf
        # https://raw.githubusercontent.com/StellarCN/scp_zh/master/fonts/SimKai.ttf
        # https://raw.githubusercontent.com/StellarCN/scp_zh/master/fonts/SimSun.ttf
        # copy to: /usr/local/lib/python3.9/site-packages/matplotlib/mpl-data/fonts/ttf/
        # use vi to edit: /usr/local/lib/python3.9/site-packages/matplotlib/mpl-data/matplotlibrc
        # add to: #font.sans-serif: SimHei,
        # then remove cache file: ~/.matplotlib/fontlist-*.json
        mpl.rcParams['font.sans-serif'] = ['SimHei']
        mpl.rcParams['font.family']='sans-serif'
        mpl.rcParams['axes.unicode_minus'] = False

        # line style
        mpl.rcParams['axes.prop_cycle'] = cycler(color=['dodgerblue', 'deeppink', 'navy', 'teal', 'maroon', 'darkorange', 'indigo'])
        mpl.rcParams['lines.linewidth'] = .5

        # China market candle color, red for up, green for down
        mc = mpf.make_marketcolors(
            up = 'red' if red_up else 'green',
            down = 'green' if red_up else 'red',
            edge = 'i',
            wick = 'i',
            volume = 'in',
            inherit = True
        )
        style = mpf.make_mpf_style(
            base_mpl_style = "ggplot",
            marketcolors = mc,
        )

        # we have panel 0 for main, panel 1 for volume by default
        next_panel = 2
        more_plot = []

        df = self.daily_df

        panel_ratios = [1, 0.3]
        keys = [
            'ma5', 'ma10', 'ma20', 'ma30', 'ma60',
            'ema5', 'ema10', 'ema20', 'ema30', 'ema60',
            'hma5', 'hma10', 'hma20', 'hma30', 'hma60',
        ]
        cols = list(set(keys) & set(df.columns))
        if len(cols) > 0:
            more_plot.append(mpf.make_addplot(df[cols], panel=0))

        if 'macd_hist' in df.columns:
            macd = df.macd_hist
            macd_color = ['r' if v >= 0 else 'g' for v in macd]
            more_plot.append(mpf.make_addplot(macd, type='bar', width=0.3, panel=next_panel, color=macd_color, ylabel='MACD'))
            cols = []
            for k in ['macd_dif', 'macd_dea']:
                if k in df.columns:
                    cols.append(k)
            if len(cols) > 0:
                more_plot.append(mpf.make_addplot(df[cols], panel=next_panel, secondary_y=False))
            next_panel = next_panel +1
            panel_ratios.append(0.3)

        indicators = get_all_signal_indicators()
        for key, values in indicators.items():
            if key in ['ma', 'macd']:
                continue
            ind_label = values['label']
            ind_cols = values['cols']

            if ind_cols[0] in df.columns:
                cols = list(set(ind_cols) & set(df.columns))
                more_plot.append(mpf.make_addplot(df[cols], panel=next_panel, ylabel=ind_label, secondary_y=False))
                next_panel = next_panel +1
                panel_ratios.append(0.3)

        if 'trade_signal' in df.columns:
            trade_signal = df['trade_signal']
            trade_color = ['r' if v >= 0 else 'g' for v in trade_signal]
            more_plot.append(mpf.make_addplot(trade_signal, type='bar', panel=next_panel, color=trade_color, ylabel='trade'))
            next_panel = next_panel +1
            panel_ratios.append(0.3)

        if 'long_pos' in df.columns:
            long_pos = df['long_pos']
            short_pos = 1 - long_pos
            more_plot.append(mpf.make_addplot(long_pos, type='bar', panel=next_panel, color='r', ylabel='position'))
            more_plot.append(mpf.make_addplot(short_pos, type='bar', panel=next_panel, color='g', secondary_y=False))
            next_panel = next_panel +1
            panel_ratios.append(0.3)

        if 'act_return' in df.columns:
            if 'daily_return' in df.columns:
                more_plot.append(mpf.make_addplot(df['daily_return'], type='line', panel=next_panel, color='b', ylabel='return'))
            more_plot.append(mpf.make_addplot(df['act_return'], type='line', panel=next_panel, color='r', secondary_y=False))
            next_panel = next_panel +1
            panel_ratios.append(0.3)

        # find the columns started with "." as performance data
        ret_cols = []
        for k in df.columns:
            if k.endswith('.'):
                ret_cols.append(k)
        if len(ret_cols) > 0:
            more_plot.append(mpf.make_addplot(df['close']/df['close'].iloc[0]-1, type='line', panel=next_panel, color='b', ylabel='cum\nreturn'))
            more_plot.append(mpf.make_addplot(df[ret_cols], type='line', panel=next_panel, secondary_y=False))
            last_panel = next_panel
            next_panel = next_panel +1
            panel_ratios.append(1.0 if (len(ret_cols)>1) else 1.0)

        kwargs = {
            'title': self.symbol + ' - ' + self.name,
            'type': "candle",
            'volume': True,
            'datetime_format': '%Y-%m-%d',
            'ylabel': "price",
            'ylabel_lower': "volume",
            'figratio': (10,6),
            'figscale': 1.0,
            'panel_ratios': panel_ratios,
            'tight_layout': True,
            'style': style,
            'addplot': more_plot,
            'returnfig': True,
            'warn_too_much_data': 244 * 5,
        }
        if out_file is not None:
            kwargs[ 'savefig' ] = out_file

        fig, axes = mpf.plot(df[['open','close','high','low','volume']], **kwargs)

        # matplotlib subplot does not support legend, we do it by ourselves
        if(len(ret_cols) > 1):
            ret_cols.insert(0, 'price')
            axes[ last_panel * 2].legend(ret_cols, loc = 'upper left')

        plt.show()




class OrderCost:
    close_tax = 0.001
    open_commission = 0.0003
    close_commission = 0.0003
    min_commission = 5.0

    def __init__(self, close_tax = 0.001, open_commission = 0.0003, close_commission = 0.0003, min_commission = 5.0):
        self.close_tax = close_tax
        self.open_commission = open_commission
        self.close_commision = close_commission
        self.min_commission = min_commission



# -*- coding: utf-8 -*-
import pandas as pd
import numpy as np
from sys import float_info as sflt

from .basic import CROSS
from ..core.indicator_signal import register_signal_indicator

# Vertical Horizontal Filter
# --------------------------------------------
def VHF(close, length = 28):
    hcp = close.rolling(length, min_periods=1).max()
    lcp = close.rolling(length, min_periods=1).min()
    sm = (close - close.shift(1)).abs().rolling(length, min_periods=1).sum()
    vhf = (hcp - lcp) / sm
    return vhf

def signal_vhf(df, inplace=False, length=28):
    vhf = VHF(df.close)
    if inplace:
        df['vhf'], df['vhf_lower'] = vhf, 0.35
    return CROSS(vhf, 0.35)

register_signal_indicator('vhf', signal_vhf, ['vhf', 'vhf_lower'], 'VHF', 'trend')



# -*- coding: utf-8; py-indent-offset:4 -*-

import datetime as dt
import pandas as pd

from ..utils import datetime_today, dict_from_df
from .data_cache import *
from .order_cost import *

class Market:
    adjust = 'hfq'

    all_symbol_name = {}
    watching_symbols = []

    symbol_daily = {}
    symbol_adjust_factor = {}
    symbol_daily_adjusted = {}

    verbose = False

    date_start = None
    date_end = None
    current_date = None
    current_time = None

    force_open = False

    def __init__(self, start, end, adjust = 'hfq'):
        self.date_start = start
        self.date_end = end
        self.current_date = self.current_time = start

        self.all_symbol_name = dict_from_df(get_all_stock_list_df(), 'symbol', 'name')

        self.adjust = adjust
        self.last_spot_time = dt.datetime.now() - dt.timedelta(minutes =10)

    def set_verbose(self, verbose = True):
        self.verbose = verbose

    # for testing purpose
    def set_force_open(self, force_open = True):
        self.force_open = force_open

    # for testing purpose
    def set_date_range(self, start, end):
        self.date_start = start
        self.date_end = end
        self.current_date = self.current_time = start

    def is_open(self):
        if self.force_open:
            return True

        now = dt.datetime.now()
        wday = now.weekday() +1
        # monday to friday, 9:00 - 15:00 treat as open time
        return (wday>=1) and (wday<=5) and (now.hour>=9 and now.hour<15)

    def load_history_price(self, symbol):
        if symbol.startswith('sh') or symbol.startswith('sz'):
            df = get_daily( symbol )
            self.symbol_daily[ symbol ] = df
            self.symbol_daily_adjusted[ symbol ] = df
        else:
            df = get_daily( symbol )
            factor_df = get_daily_adjust_factor( symbol )

            self.symbol_daily[ symbol ] = df
            self.symbol_adjust_factor[ symbol ] = factor_df

            if (self.adjust == 'hfq' or self.adjust == 'qfq'):
                self.symbol_daily_adjusted[ symbol ] = adjust_daily_with_factor(df, factor_df, self.adjust)
            else:
                self.symbol_daily_adjusted[ symbol ] = df

    def watch(self, symbols):
        for symbol in symbols:
            if not (symbol in self.watching_symbols):
                self.watching_symbols.append(symbol)
                self.load_history_price(symbol)

    def keep_daily_uptodate(self):
        for symbol in self.watching_symbols:
            self.load_history_price( symbol )

    def update_daily_realtime(self, verbose = False):
        if not self.is_open():
            return

        now = dt.datetime.now()
        data_updated = False

        # if the spot out of date, update for all stocks under watch
        next_spot_time = self.last_spot_time + dt.timedelta(seconds=10)
        if now > next_spot_time:
            spot_df = get_stock_spot(self.watching_symbols, verbose)
            if verbose:
                print('')
                print(spot_df)
            self.last_spot_time = now

            today = datetime_today()
            for i, spot_row in spot_df.iterrows():
                symbol = spot_row['symbol']
                adjust_factor = self.symbol_adjust_factor[ symbol ]['factor'].iloc[-1]
                spot_date = spot_row['date']

                data_updated = (spot_date >= today)

                new_row = self.symbol_daily[ symbol ].iloc[-1].copy()
                for k in ['open', 'high', 'low', 'close', 'volume']:
                    new_row[ k ] = spot_row[ k ]
                self.symbol_daily[ symbol ].loc[spot_date] = new_row

                new_row_adjusted = new_row.copy()
                if (self.adjust == 'hfq' or self.adjust == 'qfq'):
                    for k in ['open', 'high', 'low', 'close']:
                        new_row_adjusted[ k ] *= adjust_factor
                self.symbol_daily_adjusted[ symbol ].loc[spot_date] = new_row_adjusted

        return data_updated

    def get_index_daily(self, symbol, start = None, end = None, count = None):
        if end is None:
            end = self.date_end
        if start is None:
            start = self.date_start - dt.timedelta(days=90)
        df = get_index_daily( symbol )

        if start:
            df = df[ pd.to_datetime(start) : ]
        if end:
            df = df[ : pd.to_datetime(end) ]
        if count:
            df = df.iloc[ -count : ]

        return df

    def get_daily(self, symbol, start = None, end = None, count = None):
        if end is None:
            end = self.date_end
        if start is None:
            start = self.date_start - dt.timedelta(days=90)

        if not symbol in self.watching_symbols:
            self.watch([ symbol ])
            if end and (end >= datetime_today()):
                self.update_daily_realtime()

        df = self.symbol_daily_adjusted[ symbol ]

        if start:
            df = df[ pd.to_datetime(start) : ]
        if end:
            df = df[ : pd.to_datetime(end) ]
        if count:
            df = df.iloc[ -count : ]

        return df

    def get_spot(self, symbol, date = None):
        if date is None:
            date = self.current_date
        df = self.get_daily(symbol, end = date, count = 1)
        if df.shape[0] > 0:
            return df.iloc[-1]
        else:
            return False

    def get_price(self, symbol, date = None):
        if date is None:
            date = self.current_date
        df = self.get_daily(symbol, end = date, count = 1)
        if df.shape[0] > 0:
            return df.iloc[-1]['close']
        else:
            print(date, 'not trading day')
            return 0

    def get_real_price(self, symbol, date = None):
        if date is None:
            date = self.current_date
        if not symbol in self.watching_symbols:
            self.watch([ symbol ])
            today = datetime_today()
            if date and (date >= today):
                self.update_daily_realtime()
        df = self.symbol_daily[ symbol ]
        df = df[:date]
        if df.shape[0] > 0:
            return df['close'].iloc[-1]
        else:
            print(date, 'not trading day')
            return 0

    def get_adjust_factor(self, symbol, date = None):
        if date is None:
            date = self.current_date
        if not symbol in self.watching_symbols:
            self.watch([ symbol ])
            today = datetime_today()
            if date and (date >= today):
                self.update_daily_realtime()
        factor_df = self.symbol_adjust_factor[ symbol ]
        df = factor_df[:date]
        factor =  df['factor'].iloc[-1] if (df.shape[0] > 0) else False
        if not factor:
            print(factor_df)
        return factor

    def get_name(self, symbol):
        if symbol in self.all_symbol_name:
            return self.all_symbol_name[ symbol ]
        else:
            return ''

    def allow_trading(self, symbol, date = None):
        if date is None:
            date = self.current_date
        if not symbol in self.watching_symbols:
            self.watch([ symbol ])
            today = datetime_today()
            if date >= today:
                self.update_daily_realtime()

        df = self.symbol_daily_adjusted[ symbol ]
        return date in df.index

    def get_all_trading_dates(self):
        date_set = ()
        for symbol, df in self.symbol_daily.items():
            date_set |= set(df.index)
        dates = list(date_set)
        dates.sort()
        return dates



#!/usr/bin/env python
# -*- coding: utf-8; py-indent-offset:4 -*-
from __future__ import (absolute_import, division, print_function,
                        unicode_literals)

import sys
import os
sys.path.append(os.getcwd())

import configparser

from ..version import __version__
from ..utils import parse_params_options, dict_from_config_items
from ..core import set_lang

from .cli_create import cli_create
from .cli_finance import cli_finance
from .cli_pepb import cli_pepb
from .cli_stockpool import cli_stockpool
from .cli_strategy import cli_strategy
from .cli_fund import cli_fund

from .cli_stock import cli_stock
from .cli_index import cli_index
from .cli_pattern import cli_pattern
from .cli_indicator import cli_indicator

def cli_main_params_options(params, options):
    # parse command line arguments
    syntax_tips = '''Syntax:
    __argv0__ <command> [options]

Global Commands:
    help ................... Get help for a command
    create ................. Create a hiquant project

Project Commands:
    list ................... List stock, index, indicator, pattern, strategy

    finance ................ Finance analysis on stock pool
    pepb ................... PE/PE analysis on stock pool
    symbol ................. Create or edit symbol list .csv file
    strategy ............... Create or bench a strategy
    fund ................... Back trade or run real-time one or multiple funds

    stock .................. Plot a stock with indicators
    index .................. Plot index with indicators
    pattern ................ Draw K-line patterns
    indicator .............. List indicators

Options:
    -v, --version .......... print out the version
    -d, --verbose .......... debug mode produces verbose log output

Example:
    __argv0__ create myFund
    cd myFund

    __argv0__ list stock
    __argv0__ list index

    __argv0__ stockpool create my_stocks.csv 600036 000002 600276 300357
    __argv0__ finance view my_stocks.csv
    __argv0__ pepb view my_stocks.csv

    __argv0__ stock 600036 -ma -macd -kdj

    __argv0__ strategy create my_strategy.py
    ... modify my_strategy.py ...

    __argv0__ fund create my_fund.conf my_strategy.py
    ... modify my_fund.conf ...

    __argv0__ fund backtrade my_fund.conf

    __argv0__ fund run my_fund.conf
'''.replace('__argv0__',os.path.basename(sys.argv[0]))

    cli_tools = {
        'create': cli_create,
        'finance': cli_finance,
        'pepb': cli_pepb,
        'stockpool': cli_stockpool,
        'strategy': cli_strategy,
        'fund': cli_fund,
        'stock': cli_stock,
        'index': cli_index,
        'indicator': cli_indicator,
        'pattern': cli_pattern,
    }

    if ('-v' in options) or ('--version' in options):
        print( __version__ )
        return

    if len(params) == 0:
        print( syntax_tips )
        return

    config_file = 'hiquant.conf'
    if os.path.isfile(config_file):
        print( 'reading config from from:', config_file)
        config = configparser.ConfigParser()
        config.read(config_file, encoding='utf-8')
        main_conf = dict_from_config_items(config.items('main'))
        if 'lang' in main_conf:
            set_lang(main_conf['lang'])

    command, params = params[0], params[1:]

    if command == 'help':
        if (len(params) > 0) and (params[0] in cli_tools):
                func = cli_tools[ params[0] ]
                func(['help'], options)
        else:
            print( syntax_tips )

    elif command == 'list':
        tools_with_list = ['stock', 'index', 'indicator', 'pattern', 'strategy']
        if (len(params) > 0) and (params[0] in tools_with_list):
            func = cli_tools[ params[0] ]
            func(['list'], options)
        else:
            print('''Syntax:
    __argv0__ list [stock | index | indicator | pattern | strategy]
'''.replace('__argv0__',os.path.basename(sys.argv[0])))

    elif command in cli_tools:
        func = cli_tools[ command ]
        func(params, options)
    else:
        print( syntax_tips )

    print('')

def cli_main():
    params, options = parse_params_options(sys.argv)
    cli_main_params_options(params, options)

if __name__ == "__main__":
    cli_main()



# -*- coding: utf-8; py-indent-offset:4 -*-
from __future__ import (absolute_import, division, print_function, unicode_literals)

import os
import sys

__version__ = '0.2.621'

__hqversion__ = tuple(int(x) for x in __version__.split('.'))

if __name__ == "__main__":
    if len(sys.argv) > 1 and sys.argv[1] == 'bump':
        fp = open(__file__, 'r')
        lines = fp.readlines()
        fp.close()

        text = ''
        for line in lines:
            if line.startswith('__version__'):
                items = line.split("'")
                current_version = items[1]
                vers = items[1].split('.')
                vers[-1] = str(int(vers[-1]) +1)
                new_version = items[1] = '.'.join(vers)
                line = "'".join(items)
            text += line

        fp = open(__file__, 'w')
        fp.write(text)
        fp.close()
        print('Current version:', current_version)
        print('Bumped to:', new_version)
        print('File updated:', __file__, '\n')



# -*- coding: utf-8 -*-

from .basic import CROSS, NON_ZERO, REF
from .ma import RMA
from ..core.indicator_signal import register_signal_indicator

# --------------------------------------------
def _rsi(up, dn, length):
    # Notice!!
    # It's not correct to use SUM or SMA as mentioned in many website
    # should use EMA instead.
    # a = SUM(up, period)
    # b = SUM(dn, period)
    # see: https://zhuanlan.zhihu.com/p/351213302
    a = RMA(up, length)
    b = RMA(dn, length)
    return a / NON_ZERO(a + b) * 100.0

# Relative Strength Index, by Welles Wilder, 1978
# --------------------------------------------
def RSI(close, fast = 6, middle = 12, slow = 24):
    up = close - REF(close, 1)
    dn = up.copy()
    up[ up < 0 ] = 0
    dn[ dn > 0 ] = 0
    return _rsi(up,dn,fast), _rsi(up,dn,middle), _rsi(up,dn,slow)

def signal_rsi(df, inplace=False, fast=6, middle=12, slow=24):
    rfast, rmid, rslow = RSI(df.close, fast=fast, middle=middle, slow=slow)
    if inplace:
        df['rsi_fast'], df['rsi_middle'], df['rsi_slow'] = rfast, rmid, rslow
        df['rsi_upper'], df['rsi_lower'] = 70, 30
    cross = CROSS(rfast, rmid)
    cross_up = CROSS(rfast, 70)
    cross_dn = CROSS(rfast, 30)
    buy_signal = (cross > 0) | (cross_dn < 0)
    sell_signal = (cross < 0) | (cross_up > 0)
    return buy_signal.astype(int) - sell_signal.astype(int)

register_signal_indicator('rsi', signal_rsi, ['rsi_fast', 'rsi_middle', 'rsi_slow', 'rsi_upper', 'rsi_lower'], 'RSI', 'obos')



# -*- coding: utf-8; py-indent-offset:4 -*-

import pandas as pd
import datetime as dt
import time
import matplotlib.pyplot as plt

from ..utils import *
from .stock_market import *

class Callback:
    context = None
    func = None
    called = False
    def __init__(self, func, context):
        self.context = context
        self.func = func
        self.called = False

class Trader:
    market = None
    date_start = None
    date_end = None

    timer_callbacks = {}
    bar_callbacks = []
    funds = []

    verbose = False

    def __init__(self, market):
        self.market = market

    def set_verbose(self, verbose = True):
        self.verbose = verbose
        for fund in self.funds:
            fund.set_verbose(verbose)

    def run_daily(self, func, context, time='open'):
        # map to time
        _run_time = {
            'before_open': '7:00',
            'open': '9:30',
            'after_open': '10:00',
            'before_close': '14:30',
            'after_close': '16:00',
        }
        for k in _run_time:
            if time == k:
                time = _run_time[k]

        if time in self.timer_callbacks:
            callback_list = self.timer_callbacks[ time ]
            callback_list.append( Callback(func, context) )
        else:
            self.timer_callbacks[ time ] = [ Callback(func, context) ]

    def run_on_bar_update(self, func, context):
        self.bar_callbacks.append( Callback(func, context) )

    def add_fund(self, fund):
        self.funds.append(fund)

    def run_fund(self, date_start, date_end, tick_period = 300):
        market = self.market
        self.date_start = date_start
        self.date_end = date_end

        market.current_time = market.current_date = date = date_start
        for fund in self.funds:
            fund.init_fund( date_end >= datetime_today() )

        while date < date_end:
            # print the date at bottom of screen, but do not change line
            #if not fund.verbose:
            print('\r... {} ...'.format(date.strftime('%Y-%m-%d %H:%M:%S')), end='', flush=True)

            next_date = min(date + dt.timedelta(days=1), date_end)

            for fund in self.funds:
                fund.before_day()

            # init timer for the day, convert str to real time value
            timers = {}
            for k in self.timer_callbacks:
                time_only = dt.datetime.strptime(k, '%H:%M')
                call_time = dt.datetime(date.year, date.month, date.day, time_only.hour, time_only.minute)
                callback_list = self.timer_callbacks[ k ]
                timers[ call_time ] = callback_list
                for callback in callback_list:
                    callback.called = False

            # if the day is in the past, it's a back-trade test
            if date < datetime_today():
                timer_keys = list(timers.keys())
                timer_keys.sort()
                for k in timer_keys:
                    for fund in self.funds:
                        fund.before_tick()

                    market.current_time = k
                    callback_list = timers[k]
                    for callback in callback_list:
                        if callback.called:
                            continue
                        func = callback.func
                        func( callback.context )
                        callback.called = True

                    for fund in self.funds:
                        fund.after_tick()
            else:
                # date == today, it's realtime trade
                # wait until this day really begin
                while dt.datetime.now() < date:
                    time.sleep(1)

                # download the daily at idle time
                market.keep_daily_uptodate()

                market.current_time = dt.datetime.now()
                next_time_tick = market.current_time

                # loop until end of this day
                while market.current_time < next_date:
                    # if time to call tick
                    if market.current_time >= next_time_tick:
                        for fund in self.funds:
                            fund.before_tick()

                        if market.is_open():
                            # fetching the realtime price / volume data
                            data_updated = market.update_daily_realtime( market.verbose )
                            if data_updated:
                                # callback on bar update
                                for callback in self.bar_callbacks:
                                    func = callback.func
                                    func( callback.context )

                        for fund in self.funds:
                            fund.after_tick()

                        next_time_tick += dt.timedelta(seconds = tick_period)

                    # callback triggered by timer
                    for k in timers:
                        if market.current_time >= k:
                            callback_list = timers[k]
                            for callback in callback_list:
                                if callback.called:
                                    break
                                func = callback.func
                                func( callback.context )
                                callback.called = True

                    now_str = market.current_time.strftime('%Y-%m-%d %H:%M:%S')
                    print('\r... {} ...'.format(now_str), end = '', flush=True)
                    time.sleep(1)

                    market.current_time = dt.datetime.now()

            for fund in self.funds:
                fund.after_day()

            # move to next date
            market.current_date = date = next_date

        print('')
        pass

    def get_report(self):
        report_df = pd.DataFrame()
        for fund in self.funds:
            row = fund.get_summary()
            if report_df.shape[0] == 0:
                report_df = pd.DataFrame([], columns = list(row.keys()))
            report_df = report_df.append(row, ignore_index = True)
        return report_df

    def print_report(self):
        report_df = self.get_report()

        report_df.set_index('fund_id', inplace=True, drop=True)
        report_df = report_df.T
        report_df.index += ':'
        report_df.index.set_names('', inplace=True)
        report_df.columns.set_names('', inplace=True)
        report_df.index = report_df.index.str.pad(15, side='left')

        print('-' * 80)
        print(report_df)
        print('-' * 80)

    def plot(self, compare_index = None, out_file= None):
        df = pd.DataFrame([], index=pd.date_range(start = self.date_start, end = self.date_end))
        for fund in self.funds:
            stat_df = fund.get_stat()

            # if only one strategy, we also plot the buy/sell and drawdown
            if len(self.funds) == 1:
                df = stat_df[['buy', 'sell', 'drawdown']]

            value = stat_df['value']
            df[ fund.fund_name ] = (value / value.iloc[0] - 1) * 100.0

        if compare_index:
            index_name = dict_from_df(get_all_index_list_df(), 'symbol', 'name')
            compare_index_name = index_name[ compare_index ] if (compare_index in index_name) else compare_index

            cmp_value = self.market.get_index_daily(compare_index, start=self.date_start, end=self.date_end)['close']
            cmp_value = (cmp_value / cmp_value.iloc[0] -1) * 100.0

            df[ compare_index_name ] = cmp_value

        df.dropna(inplace=True)
        df.index = df.index.strftime('%Y-%m-%d')

        # now plot to visualize
        if len(self.funds) > 1:
            # if multi strategy, we compare the performance
            df.plot(figsize = (10,6), grid = True, xlabel = 'date', ylabel = 'return (%)', title = 'performance')
            plt.show()
        else:
            # if only one strategy, we also plot the buy/sell and drawdown
            fig, axes = plt.subplots(nrows=3, gridspec_kw={'height_ratios': [3, 1, 1]})

            fund_name = self.funds[0].fund_name
            df[[fund_name, compare_index_name]].plot(ax=axes[0], figsize = (10,6), grid = True, sharex=axes[0], label = 'date', ylabel = 'return (%)', title = 'performance')

            axes[1].bar(df.index, df.buy, color='r')
            axes[1].bar(df.index, df.sell, color='g')
            axes[1].set_ylabel('trade count')

            df['drawdown'] = df['drawdown'] * 100.0
            df[['drawdown']].plot(ax=axes[2], grid = True, sharex=axes[0], ylabel = 'drawdown (%)', legend=False)

            plt.xticks(rotation=45)

            if out_file is not None:
                plt.savefig(out_file)
            else:
                plt.show()



# -*- coding: utf-8; py-indent-offset:4 -*-

from hiquant.utils.args import dict_from_config_items
import os
import sys
from tabulate import tabulate
import configparser

from ..core import list_signal_indicators, dict_from_df, get_all_stock_list_df
from ..core import date_from_str, Market, Stock, OrderCost

def cli_stock(params, options):
    syntax_tips = '''Syntax:
    __argv0__ stock list [update]
    __argv0__ stock <symbol> [<from_date>] [<to_date>] [<options>]

<symbol> ....................... stock symbol, like 600036
<from_date> .................... default from 3 year ago
<to_date> ...................... default to yesterday

<options> ...................... options to plot the stock
    -ma, -macd, -kdj, etc. ..... add some indicators
    -all ....................... add all indicators
    -mix ....................... show trade result if mix trade signal of the indicators
    -top5 ...................... show only top 5 indicators according to result

Example:
    __argv0__ stock 600036 -ma
    __argv0__ stock 600036 20180101 -ma -macd -kdj
    __argv0__ stock 600036 20180101 20210101 -ma -macd -kdj
    __argv0__ stock 600036 -ma -macd -mix
'''.replace('__argv0__',os.path.basename(sys.argv[0]))

    if (len(params) == 0) or (params[0] == 'help'):
        print( syntax_tips )
        return

    if params[0] == 'list':
        force_update = False
        if len(params) > 1 and (params[1] == 'update'):
            force_update = True
        if '-update' in options or '-u' in options:
            force_update = True
        df = get_all_stock_list_df(force_update= force_update)
        print( tabulate(df, headers='keys', showindex=False, tablefmt='psql') )
        print( 'Totally', df.shape[0], 'rows.\n')
        return

    symbol_name = dict_from_df(get_all_stock_list_df(), 'symbol', 'name')

    symbol = params[0]
    if symbol in symbol_name:
        name = symbol_name[ symbol ]
    else:
        print('\nError: symbol not found in stock list:', symbol)
        return

    date_start = date_from_str(params[1] if len(params) > 1 else '3 years ago')
    date_end = date_from_str(params[2] if len(params) > 2 else 'yesterday')

    market = Market(date_start, date_end)
    df = market.get_daily(symbol, start = date_start, end = date_end)
    stock = Stock(symbol, name, df)

    all_indicators = list_signal_indicators()

    indicators = []
    topN = 5
    inplace = True
    out_file = None
    for option in options:
        if option.startswith('-top'):
            topN = int(option.replace('-top',''))
        elif option.startswith('-out=') and (option.endswith('.png') or option.endswith('.jpg')):
            out_file = option.replace('-out=', '')
        else:
            k = option.replace('-','')
            if (k in all_indicators) and (not k in indicators):
                indicators.append(k)
    if '-all' in options:
        indicators = all_indicators
        inplace = False

    mix = '-mix' in options

    config_file = 'hiquant.conf'
    if os.path.isfile(config_file):
        print( 'reading config from from:', config_file)
        config = configparser.ConfigParser()
        config.read(config_file, encoding='utf-8')
        order_cost_conf = dict_from_config_items(config.items('order_cost'))
        order_cost = OrderCost(
            float(order_cost_conf['close_tax']),
            float(order_cost_conf['open_commission']),
            float(order_cost_conf['close_commission']),
            float(order_cost_conf['min_commission']),
        )
    else:
        order_cost = OrderCost(0.001, 0.0003, 0.0003, 5.0)

    # add indicators and calculate performance
    stock.add_indicator(indicators, mix=mix, inplace= inplace, order_cost = order_cost)

    rank_df = stock.rank_indicator(by = 'final')
    if rank_df.shape[0] > 0:
        print(rank_df)

    # find top indicators
    ranked_indicators = rank_df['indicator'].tolist()
    other_indicators = ranked_indicators[topN:]
    df.drop(columns = other_indicators, inplace=True)

    stock.plot(out_file= out_file)



# -*- coding: utf-8 -*-

from ..core.lang import add_lang

add_lang('en', {
    'buy': 'buy',
    'sell': 'sell',
    'stop loss': 'stop loss',
    'stop earn': 'stop earn',
    'signal': 'signal',
})




import pandas as pd

from .stock import *

class Portfolio:
    market = None
    available_cash = 0.0
    positions = {}
    history = {}

    def __init__(self, market):
        self.market = market
        self.available_cash = 0.0
        self.positions = {}
        self.history = {}

    def total_value(self):
        market = self.market
        value = self.available_cash
        for symbol, stock in self.positions.items():
            value += stock.shares * market.get_price(symbol)
        return value

    def to_dataframe(self, use_real_price = True):
        market = self.market

        table = [['cash', '-', self.available_cash, 1.0]]
        for symbol, stock in self.positions.items():
            shares = stock.shares
            cost = stock.cost
            if use_real_price:
                adjust_factor = 1.0 / market.get_adjust_factor(symbol)
                cost = round(cost * adjust_factor, 3)
                shares = int(shares / adjust_factor)
            row = [symbol, stock.name, shares, cost]
            table.append(row)

        return pd.DataFrame(table,columns=['symbol','name','amount','cost'])

    def from_dataframe(self, df, use_real_price = True):
        market = self.market

        self.available_cash = 0
        self.positions = {}
        self.history = {}
        for i, row in df.iterrows():
            symbol = row['symbol']
            if symbol == 'cash':
                self.available_cash = float(row['amount'])
            else:
                name = row['name']
                cost = float(row['cost'])
                shares = float(row['amount'])

                if use_real_price:
                    adjust_factor = market.get_adjust_factor(symbol)
                    cost *= adjust_factor
                    shares /= adjust_factor

                stock = Stock(symbol, name)
                stock.shares = shares
                stock.cost = cost
                self.positions[ symbol ] = stock

    def to_csv(self, csvfile, use_real_price = True):
        df = self.to_dataframe(use_real_price)
        df.to_csv(csvfile, index=False)

    def from_csv(self, csvfile, use_real_price = True):
        df = pd.read_csv(csvfile, dtype=str)
        self.from_dataframe(df, use_real_price)




import os
import datetime as dt
import pandas as pd

from .data_source_akshare import *
from ..utils import *

def get_cached_download_df(csv_file, download_func, param = None, check_date = False):
    if type(param) == str:
        csv_file = csv_file.replace('{param}', param)

    need_update = False
    if os.path.isfile(csv_file):
        if check_date:
            modified_time = get_file_modify_time(csv_file)
            need_update = modified_time < datetime_today()
        else:
            need_update = False
    else:
        need_update = True

    if need_update:
        df = download_func(param)
        df.to_csv(csv_file, index= bool(df.index.name))
    else:
        df = pd.read_csv(csv_file, dtype=str)

    return df

def get_all_stock_list_df(force_update= False):
    return get_cached_download_df('cache/all_stock_list.csv', download_stock_list, check_date= force_update)

def get_all_index_list_df(force_update= False):
    return get_cached_download_df('cache/all_index_list.csv', download_index_list, check_date= force_update)

def get_all_stock_symbol_name():
    return dict_from_df(get_all_stock_list_df(), 'symbol', 'name')

def get_stockpool_df(symbols):
    if type(symbols) == str:
        if symbols.endswith('.csv'):
            return pd.read_csv(symbols, dtype=str)
        else:
            symbols = symbols.replace(' ','').split(',')

    df = get_all_stock_list_df()
    all_symbols = df['symbol'].tolist()
    invalid_symbols = list(set(symbols) - set(all_symbols))
    if len(invalid_symbols) > 0:
        raise ValueError('Not found in stock list. Invalid symbols: ' + ', '.join(invalid_symbols))

    return df[ df['symbol'].isin(symbols) ].reset_index(drop=True)

def get_index_daily( symbol ):
    df = get_cached_download_df('cache/market/{param}_daily.csv', download_func= download_index_daily, param= symbol, check_date= True)

    if 'date' in df.columns:
        df['date'] = pd.to_datetime(df['date'])
        df.set_index('date', inplace= True, drop= True)
        df = df.astype(float)

    return df

def get_daily( symbol ):
    df = get_cached_download_df('cache/market/{param}_daily.csv', download_func= download_stock_daily, param= symbol, check_date= True)

    if 'date' in df.columns:
        df['date'] = pd.to_datetime(df['date'])
        df.set_index('date', inplace= True, drop= True)
        df = df.astype(float)

    return df

def get_daily_adjust_factor( symbol ):
    df = get_cached_download_df('cache/market/{param}_daily_factor.csv', download_func= download_stock_daily_adjust_factor, param= symbol, check_date= True)

    if 'date' in df.columns:
        df['date'] = pd.to_datetime(df['date'])
        df.set_index('date', inplace= True, drop= True)
        df = df.astype(float)

    return df

def adjust_daily_with_factor(daily_df, factor_df, adjust: str = 'hfq'):
    if adjust == 'qfq':
        factor_df = factor_df.copy()
        max_date_factor = factor_df.iloc[-1]['factor']
        factor_df['factor'] = factor_df['factor'] / max_date_factor

    if 'factor' in daily_df.columns:
        del daily_df['factor']

    daily_df = pd.merge(
        daily_df, factor_df, left_index=True, right_index=True, how="outer"
    )
    daily_df.fillna(method="ffill", inplace=True)
    daily_df = daily_df.astype(float)
    daily_df.dropna(inplace=True)
    daily_df.drop_duplicates(subset=["open", "high", "low", "close", "volume"], inplace=True)
    daily_df["open"] = daily_df["open"] * daily_df["factor"]
    daily_df["high"] = daily_df["high"] * daily_df["factor"]
    daily_df["close"] = daily_df["close"] * daily_df["factor"]
    daily_df["low"] = daily_df["low"] * daily_df["factor"]
    daily_df["open"] = round(daily_df["open"], 2)
    daily_df["high"] = round(daily_df["high"], 2)
    daily_df["low"] = round(daily_df["low"], 2)
    daily_df["close"] = round(daily_df["close"], 2)
    daily_df.dropna(inplace=True)

    return daily_df

def adjust_daily(symbol, daily_df, adjust: str = 'hfq'):
    factor_df = get_daily_adjust_factor( symbol )
    return adjust_daily_with_factor(daily_df, factor_df, adjust)

def get_stock_spot(symbols, verbose = False):
    return download_stock_spot(symbols, verbose)

def get_finance_balance_report(symbol, force_update= False):
    return get_cached_download_df('cache/finance/{param}_balance.csv', download_finance_balance_report, symbol, check_date= force_update)

def get_finance_income_report(symbol, force_update= False):
    return get_cached_download_df('cache/finance/{param}_income.csv', download_finance_income_report, symbol, check_date= force_update)

def get_finance_cashflow_report(symbol, force_update= False):
    return get_cached_download_df('cache/finance/{param}_cashflow.csv', download_finance_cashflow_report, symbol, check_date= force_update)

def create_finance_abstract_df(symbol, force_update= False):
    balance_df = get_finance_balance_report(symbol, force_update= force_update)
    income_df = get_finance_income_report(symbol, force_update= force_update)
    cashflow_df = get_finance_cashflow_report(symbol, force_update= force_update)
    return extract_abstract_from_report(symbol, balance_df, income_df, cashflow_df)

def get_finance_abstract_df(symbol, force_update= False):
    df = get_cached_download_df('cache/finance/{param}_abstract.csv', create_finance_abstract_df, symbol, check_date= force_update)
    if 'date' in df.columns:
        df['date'] = pd.to_datetime(df['date'])
        df.set_index('date', inplace= True, drop= True)
        df = df.astype(float)
    return df

def get_ipoinfo_df(symbol, force_update= False):
    return get_cached_download_df('cache/finance/{param}_ipo.csv', download_ipo, symbol, check_date= force_update)

def get_dividend_history(symbol, force_update= False):
    return get_cached_download_df('cache/finance/{param}_dividend.csv', download_dividend_history, symbol, check_date= force_update)

def get_finance_indicator(symbol, force_update= False):
    abstract_df = get_finance_abstract_df(symbol, force_update= force_update)
    ipoinfo_df = get_ipoinfo_df(symbol, force_update= force_update)
    dividend_df = get_dividend_history(symbol, force_update= force_update)
    return extract_finance_indicator_data(symbol, abstract_df, ipoinfo_df, dividend_df)

def get_finance_indicator_df(symbols = None, force_update= False):
    symbol_name = dict_from_df(get_all_stock_list_df(), 'symbol', 'name')
    table = []
    cols = None
    i = 0
    n = len(symbols)
    for symbol in symbols:
        name = symbol_name[ symbol ]
        i += 1
        print('\r... {}/{} - {} {} ...'.format(i, n, symbol, name), end='', flush= True)
        row = get_finance_indicator(symbol, force_update= force_update)
        if cols is None:
            cols = list( row.keys() )
            cols.append('name')
        row = list(row.values())
        row.append(name)
        table.append( row )
    print('')
    return pd.DataFrame(table, columns= cols)

def get_finance_indicator_all(force_update= False):
    df = get_all_stock_list_df()
    all_symbols = df['symbol'].tolist()
    print('Processing finance indicators for all stocks ...')
    df = get_cached_download_df('cache/all_stock_indicator.csv', get_finance_indicator_df, param= all_symbols, check_date = force_update)
    df = df.astype({
        'ipo_years': 'float64',
        '3yr_grow_rate': 'float64',
        'grow_rate': 'float64',
        'grow': 'float64',
        'share_grow': 'float64',
        'start_bvps': 'float64',
        'now_bvps': 'float64',
        'eps': 'float64',
        'roe': 'float64',
        'debt_ratio': 'float64',
        'cash_ratio': 'float64',
        'earn_ttm': 'float64',
        'assets': 'float64',
        'equity': 'float64',
        'shares': 'float64',
    })
    return df

def get_stock_pepb_history(symbol, force_update= False):
    df = get_cached_download_df('cache/pepb/{param}_pepb.csv', download_stock_pepb_history, param= symbol, check_date= force_update)
    if 'date' in df.columns:
        df['date'] = pd.to_datetime(df['date'])
        df.set_index('date', inplace= True, drop= True)
        df = df.astype(float)
    return df

def get_pepb_summary(symbol, force_update= False, years = 10):
    today = datetime_today()
    years_ago = today - dt.timedelta(days= 365 * years)
    df = get_stock_pepb_history(symbol, force_update= force_update)
    if df.shape[0] > 0:
        df = df.loc[ pd.to_datetime(years_ago) : pd.to_datetime(today) ]
        df['pe_ttm'].fillna(method='ffill', inplace=True)
        df['pe_ttm'].fillna(method='bfill', inplace=True)
        df['pb'].fillna(method='ffill', inplace=True)
        df['pb'].fillna(method='bfill', inplace=True)
        pe_list = df['pe_ttm'].tolist()
        pb_list = df['pb'].tolist()
    else:
        pe_list = [ 22.0 ]
        pb_list = [ 1.0 ]

    pe = pe_list[-1]
    pb = pb_list[-1]
    pe_pos = round(sum([1 if i < pe else 0 for i in pe_list]) * 100.0 / len(pe_list), 1)
    pb_pos = round(sum([1 if i < pb else 0 for i in pb_list]) * 100.0 / len(pb_list), 1)
    mean_pe = sum(pe_list) / len(pe_list)
    mean_pb = sum(pb_list) / len(pb_list)
    pe_ratio = round(pe / mean_pe, 3)
    pb_ratio = round(pe / mean_pb, 3)

    return {
        'symbol': symbol,
        'pe': pe,
        'pb': pb,
        'pe_pos': pe_pos,
        'pb_pos': pb_pos,
        'pe_ratio': pe_ratio,
        'pb_ratio': pb_ratio,
    }

def get_pepb_symmary_df(symbols, force_update= False, years = 10):
    symbol_name = get_all_stock_symbol_name()
    table = []
    cols = None
    for symbol in symbols:
        name = symbol_name[ symbol ]
        row = get_pepb_summary(symbol, force_update= force_update, years= years)
        if cols is None:
            cols = list(row.keys())
            cols.append('name')
        row = list(row.values())
        row.append(name)
        table.append( row )
    return pd.DataFrame(table, columns=cols)

def get_pepb_symmary_all(force_update= False):
    df = get_all_stock_list_df()
    all_symbols = df['symbol'].tolist()
    print('Processing PE/PB summary for all stocks ...')
    df = get_cached_download_df('cache/all_stock_pepb.csv', get_pepb_symmary_df, symbols= all_symbols, check_date = force_update)
    df = df.astype({
        'pe': 'float64',
        'pb': 'float64',
        'pe_pos': 'float64',
        'pb_pos': 'float64',
        'pe_ratio': 'float64',
        'pb_ratio': 'float64',
    })
    return df

def get_macro_bank_interest_rate(country, force_update= False):
    df = get_cached_download_df('cache/bank/{paran}_macro_interest_rate.csv', download_macro_bank_interest_rate, param= country, check_date = force_update)
    df['date'] = pd.to_datetime(df.index.date)
    df.set_index('date', inplace=True, drop=True)
    return df




import time
import requests
import datetime as dt
import pandas as pd
import akshare as ak

from ..utils import *

def download_stock_list(param, verbose = False):
    df = ak.stock_info_a_code_name()
    time.sleep(2)
    df.columns = ['symbol', 'name']
    if verbose:
        print(df)
    return df

def download_index_list(param, verbose = False):
    df = ak.stock_zh_index_spot()
    time.sleep(2)
    df.columns = ['symbol', 'name', 'close', 'amount_diff', 'volume_diff', 'last_close', 'open', 'high', 'low', 'volume', 'amount']
    df = df[['symbol', 'name']]
    if verbose:
        print(df)
    return df

def download_index_daily( symbol ):
    daily_df = ak.stock_zh_index_daily(symbol = symbol)
    time.sleep(2)
    daily_df['date'] = pd.to_datetime(daily_df.index.date)
    daily_df.set_index('date', inplace=True, drop=True)
    return daily_df

def download_stock_daily( symbol, adjust = '' ):
    start = '20000101'
    end = dt.datetime.now().strftime('%Y%m%d')

    print('    Fetching stock daily history {} {} ...'.format(symbol, adjust))
    if symbol.startswith('hk') or symbol.startswith('HK'):
        symbol = '0' + symbol[-4:]
        df = ak.stock_hk_daily(symbol=symbol, adjust = adjust)
        time.sleep(3)
        df.index = df.index.set_names('date')
    else:
        if symbol.startswith('6'):
            symbol = 'sh' + symbol
        elif symbol.startswith('0') or symbol.startswith('3'):
            symbol = 'sz' + symbol
        df = ak.stock_zh_a_daily(symbol=symbol, start_date=start, end_date=end, adjust = adjust)
        time.sleep(3)

    if 'date' in df.columns:
        df['date'] = pd.to_datetime(df['date'])
        df.set_index('date', inplace= True, drop= True)
        df = df.astype(float)

    df.sort_index(ascending=True, inplace=True)

    # drop invalid data
    if 'close' in df.columns:
        df = df[df['close'] > 0.01]

    return df

def download_stock_daily_adjust_factor( symbol ):
    df = download_stock_daily( symbol, adjust = 'hfq-factor')
    df.columns = ['factor']
    return df

def download_stock_spot( symbols, verbose= False ):
    if len(symbols) > 100:
        df = pd.DataFrame()
        for i in range(0, len(symbols), 100):
            page = symbols[i:i+100]
            print('-------------------- page:', int(i/100), 'size:', len(page), 'range:', i, '~', i+len(page), '--------------------')
            page_df = download_stock_spot(page)
            df = df.append(page_df, ignore_index=True)
        return df
    else:
        sina_symbols = []
        for symbol in symbols:
            if symbol.startswith('6'):
                sina_symbols.append( 'sh' + symbol )
            elif symbol.startswith('0') or symbol.startswith('3'):
                sina_symbols.append( 'sz' + symbol )
            elif symbol.startswith('hk') or symbol.startswith('HK'):
                sina_symbols.append( 'hk0' + symbol[-4:] )

        print('\r... {} | fetching OHCL ...'.format(str_now()), end = '', flush = True)

        url = "http://hq.sinajs.cn/list={0}".format(",".join(sina_symbols))
        if verbose:
            print('\n', url)

        time.sleep(3)
        r = requests.get(url)
        print('\r... {} ...'.format(str_now()) + (' ' * 40), end = '', flush = True)

        table = []
        if(r.status_code == 200):
            lines = r.text.split('\n')
            for line in lines:
                strs = line.split('=')
                if(len(strs) <2):
                    break
                row = []
                str0 = strs[0].replace('var hq_str_','')
                str1 = strs[1].replace('\"','').replace(';','')
                v = str1.split(',')
                if str0.startswith('sh') or str0.startswith('sz'):
                    # return data format:
                    # symbol = name, open, prevclose, lasttrade, high, low, buy, sell, volume, amount
                    # buy1-count, buy1-price, ...
                    # sell1-count, sell1-price, ...
                    # date, time
                    str0 = str0[-6:]
                    # symbol, name, open, prevclose, lasttrade, high, low, volume, amount, date, time
                    row = [str0, v[0], v[1], v[2], v[3], v[4], v[5], v[8], v[9], v[30], v[31]]
                elif str0.startswith('hk'):
                    # return data format:
                    # symbol = en_name, zh_name, open, prevclose, high, low, lasttrade,
                    # pricechange, changepercent, buy1-price, sell1-price, amount, volume, 0, 0, ?, ?,
                    # date, time
                    str0 = 'hk' + str0[-4:]
                    # symbol, name, open, prevclose, lasttrade, high, low, volume, amount, date, time
                    row = [str0, v[1], v[2], v[3], v[4], v[4], v[6], v[12], v[11], v[17], v[18]]
                table.append(row)
                #print(row)
        df = pd.DataFrame(table)
        df.columns = [
            'symbol',
            'name',
            'open',
            'prevclose',
            'close',
            'high',
            'low',
            'volume',
            'amount',
            'date',
            'time'
        ]
        df = df.astype({
            'open':'float64',
            'prevclose':'float64',
            'close':'float64',
            'high':'float64',
            'low':'float64',
            'volume':'float64',
            'amount':'float64',
            'date':'datetime64'
        })
        #print(df)
        return df

def download_finance_report(symbol, report_type = 'balance'):
    en_zh = {'balance':'资产负债表', 'income':'利润表', 'cashflow':'现金流量表'}
    if report_type in en_zh:
        pass
    else:
        raise ValueError('Unknown finance report type: ' + report_type)

    print('\tfetching ' + report_type + ' report ...')
    df = ak.stock_financial_report_sina(symbol, en_zh[ report_type ])
    time.sleep(1)

    df.rename(columns={'报表日期':'date'}, inplace= True)
    val = df.iloc[0]['date']
    if type(val) == str:
        if '-' in val:
            df['date'] = pd.to_datetime(df['date'], format='%Y-%m-%d')
        else:
            df['date'] = pd.to_datetime(df['date'], format='%Y%m%d')

    return df

def download_finance_balance_report(symbol):
    return download_finance_report(symbol, 'balance')

def download_finance_income_report(symbol):
    return download_finance_report(symbol, 'income')

def download_finance_cashflow_report(symbol):
    return download_finance_report(symbol, 'cashflow')

def extract_data_from_report(report_df, data_mapping):
    df = pd.DataFrame()

    cols = report_df.columns
    for key, fields in data_mapping.items():
        found = False
        for field in fields:
            if field in cols:
                df[key] = report_df[field]
                found = True
        if not found:
            raise ValueError('Column not found in report: ' + key)

    df = df.astype(float)
    df.index = pd.to_datetime( report_df['date'] )

    return df

def extract_abstract_from_report(symbol, balance_df, income_df, cashflow_df):
    df1 = extract_data_from_report(balance_df, {
        'assets': ['资产总计'],
        'debt': ['负债合计'],
        'equity': ['股东权益合计','所有者权益合计','所有者权益(或股东权益)合计'],
        'shares': ['股本','实收资本(或股本)']
    })
    df1 = df1[~df1.index.duplicated(keep='first')]
    df1 = df1[df1['shares'] > 0]

    df2 = extract_data_from_report(income_df, {
        # income sheet
        'revenue': ['一、营业收入', '一、营业总收入'],
        'cost': ['二、营业支出','二、营业总成本'],
        'earning': ['五、净利润'],
    })
    df2 = df2[~df2.index.duplicated(keep='first')]

    df3 = extract_data_from_report(cashflow_df, {
        # cashflow sheet
        'operate': ['经营活动产生的现金流量净额'],
        'invest': ['投资活动产生的现金流量净额'],
        'financing': ['筹资活动产生的现金流量净额'],
        'cash': ['六、期末现金及现金等价物余额']
    })
    df3 = df3[~df3.index.duplicated(keep='first')]

    df = pd.DataFrame()
    x1 = set(df1.index.tolist())
    x2 = set(df2.index.tolist())
    x3 = set(df3.index.tolist())
    if len(x1 ^ x2) > 0 or len(x1 ^ x3) > 0:
        xu = x1.union(x2, x3)
        info = symbol + ': some quarter report missing\n'
        diff = {
            'balance': (xu-x1),
            'income': (xu-x2),
            'cashflow': (xu-x3),
        }
        for k, v in diff.items():
            dates = []
            for date in v:
                dates.append(date.strftime('%Y-%m-%d'))
            dates.sort()
            info = info + '\t' + k + ': ' + ', '.join(dates) + '\n'
        print(info)

    df = pd.concat([df1, df2, df3], axis=1, join='inner')

    df = df.astype({
        'assets':'float',
        'debt':'float',
        'equity':'float',
        'shares':'float',
        'revenue':'float',
        'cost':'float',
        'earning':'float',
        'operate':'float',
        'invest':'float',
        'financing':'float',
        'cash':'float',
    })

    df.sort_index(ascending= True)

    return df

def download_ipo(symbol):
    df = ak.stock_ipo_info(stock= symbol)
    time.sleep(2)
    return df

def download_dividend_history(symbol):
    df = ak.stock_history_dividend_detail(indicator="分红", stock=symbol, date='')
    time.sleep(2)
    df = df[df['进度'] == '实施']
    df.index = pd.to_datetime(df['公告日期'])
    df = df.astype({
        '公告日期':'datetime64',
        #'除权除息日':'datetime64',
        #'股权登记日':'datetime64',
    })
    df.sort_index(ascending= True, inplace= True)
    return df

def download_rightissue_history(symbol):
    df = ak.stock_history_dividend_detail(indicator="配股", stock=symbol, date='')
    time.sleep(2)
    df = df[df['查看详细'] == '查看']
    df.index = pd.to_datetime(df['公告日期'])
    df = df.astype({
        '公告日期':'datetime64',
        #'除权日':'datetime64',
        #'股权登记日':'datetime64',
        #'缴款起始日':'datetime64',
        #'缴款终止日':'datetime64',
        #'配股上市日':'datetime64',
    })
    df.sort_index(ascending= True, inplace= True)
    return df

def get_share_grow(dividend_df, date_from, date_to):
    df = dividend_df.astype({
        '公告日期':'datetime64',
        '送股(股)':'float64',
        '转增(股)':'float64',
        #'派息(税前)(元)':'float64',
    })
    df = df[df['公告日期'] > date_from]
    df = df[df['公告日期'] < date_to]
    x = 1.0
    n = 0
    for i, row in df.iterrows():
        give = row['送股(股)']
        convert = row['转增(股)']
        #bonus = row['派息(税前)(元)']
        # split every 10 shares, we ignore bonus temprarily
        if give>0 or convert>0:
            n = n +1
            x = x * (1 + 0.1 * give) * (1 + 0.1 * convert)
    return x, n

def extract_finance_indicator_data(symbol, abstract_df, ipoinfo_df, dividend_df):
    df = abstract_df

    df = df[df['assets'] > 0]
    df = df[df['equity'] > 0]
    df = df[df['shares'] > 0]

    # debt ratio
    df['debt_ratio'] = round(df['debt'] / df['assets'], 3)
    # leverage
    df['leverage'] = round(df['assets'] / df['equity'], 3)
    # cash ratio
    df['cash_ratio'] = round(df['cash'] / df['assets'], 3)

    # convert earning to earn_ttm according to report month
    df['earn_ttm'] = df['earning'] / df.index.month * 12.0

    # rate of return on owner equity
    df['roe'] = round(df['earn_ttm'] / df['equity'], 3)

    # book value (owner equity) per share
    df['bvps'] = round(df['equity'] / df['shares'], 3)

    # earning per share
    df['eps'] = round(df['earn_ttm'] / df['shares'], 3)

    data = {}
    ipo_dict = dict_from_df(ipoinfo_df, 'item', 'value')
    data['symbol'] = symbol
    data['ipo_date'] = ipo_date = ipo_dict['上市日期']

    if type(ipo_date) == str and ipo_date != '--':
        ipo_date = dt.datetime.strptime(ipo_date, '%Y-%m-%d')
        data['ipo_years'] = round((dt.datetime.now() - ipo_date).days / 365, 2)

        # we only need finance report no early than 3 years before IPO
        ipo_3yr_ago = pd.to_datetime( ipo_date - dt.timedelta(days=365*3) )
        df = df[ df.index >= ipo_3yr_ago ]
    else:
        data['ipo_years'] = 0

    dates = list(df.index)
    min_date = min(dates)
    max_date = max(dates)
    data['last_report'] = max_date.strftime('%Y-%m-%d')

    ago3yr = dt.datetime(max_date.year -4, 12, 31)
    if (not ago3yr in dates) or (ago3yr < min_date):
        ago3yr = min_date
    bvps_grow_3yr = df.loc[max_date]['bvps'] / df.loc[ago3yr]['bvps']
    share_grow_3yr, share_split = get_share_grow(dividend_df, ago3yr, max_date)
    grow3yr = round(bvps_grow_3yr * share_grow_3yr, 3)
    data['3yr_grow_rate'] = round(earn_to_annual(grow3yr-1, (max_date - ago3yr).days), 3) if grow3yr > 1 else 0

    bvps_grow = df.loc[max_date]['bvps'] / df.loc[min_date]['bvps']
    share_grow, share_split = get_share_grow(dividend_df, min_date, max_date)
    grow = round(bvps_grow * share_grow, 3)
    data['grow_rate'] = round(earn_to_annual(grow-1, (max_date - min_date).days), 3) if grow > 1 else 0
    data['grow'] = grow
    data['share_grow'] = round(share_grow, 2)
    #data['split'] = share_split

    data['start_bvps'] = df.iloc[0]['bvps']
    data['now_bvps'] = df.iloc[-1]['bvps']
    data['eps'] = df.iloc[-1]['eps']

    roes = df['roe']
    data['roe'] = round(sum(roes) / len(roes), 3)

    data['debt_ratio'] = df.iloc[-1]['debt_ratio']
    data['cash_ratio'] = df.iloc[-1]['cash_ratio']

    data['earn_ttm'] = round(df.iloc[-1]['earn_ttm'] / 100000000.0, 3)

    data['assets'] = round(df.iloc[-1]['assets'] / 100000000.0, 3)
    data['equity'] = round(df.iloc[-1]['equity'] / 100000000.0, 3)
    data['shares'] = round(df.iloc[-1]['shares'] / 100000000.0, 3)

    return data

def download_stock_pepb_history(symbol):
    print('\tfetching pe/pb data ...', symbol)
    df = ak.stock_a_lg_indicator(symbol)
    time.sleep(2)
    df = df.rename(columns={"trade_date":"date"}).sort_values(by='date')
    return df

def download_macro_bank_interest_rate(country):
    funcs = {
        'china': ak.macro_bank_china_interest_rate,
        'usa': ak.macro_bank_usa_interest_rate,
        'euro': ak.macro_bank_euro_interest_rate,
    }
    if country in funcs:
        func = funcs[ country ]
        data = func()
        time.sleep(2)
        return pd.DataFrame({'date':data.index, 'rate':data.values})
    else:
        raise ValueError('Invalid country, not supported yet: ' + country)
